# coding: utf8

"""Utils for the implementation of t1-volume-tissue-segmentation pipeline."""

import os

import numpy as np
from nipype.interfaces.base import (
    File,
    InputMultiPath,
    OutputMultiPath,
    TraitedSpec,
    traits,
)
from nipype.interfaces.spm.base import SPMCommand, SPMCommandInputSpec
from nipype.utils.filemanip import filename_to_list, list_to_filename


def init_input_node(t1w):
    """Extract "sub-<participant_id>_ses-<session_label>" from <t1w> and print begin message."""
    from clinica.utils.filemanip import get_subject_id
    from clinica.utils.ux import print_begin_image

    subject_id = get_subject_id(t1w)
    print_begin_image(subject_id)

    return subject_id, t1w


def print_end_pipeline(subject_id, final_file):
    """Display end message for <subject_id> when <final_file> is connected."""
    from clinica.utils.ux import print_end_image

    print_end_image(subject_id)


def zip_list_files(class_images, zip_files=False):
    from clinica.utils.filemanip import zip_nii

    if zip_files:
        return [zip_nii(tissue, True) for tissue in class_images]

    return [tissue for tissue in class_images]


def get_tissue_tuples(
    tissue_map,
    tissue_classes,
    dartel_tissues,
    save_warped_unmodulated,
    save_warped_modulated,
):
    """Get tissue tuples.

    Method to obtain the list of tuples, one for each tissue class, with the following fields:
     - tissue probability map (4D), 1-based index to frame
     - number of gaussians
     - which maps to save [Native, DARTEL] - a tuple of two boolean values
     - which maps to save [Unmodulated, Modulated] - a tuple of two boolean values

    Args:
        tissue_map: Path to tissue maps
        tissue_classes: Classes of images to obtain from segmentation. Ex: [1,2,3] is GM, WM and CSF
        dartel_tissues: Classes of images to save for DARTEL template calculation. Ex: [1] is only GM'
        save_warped_unmodulated: Save warped unmodulated images for tissues specified in --tissue_classes
        save_warped_modulated: Save warped modulated images for tissues specified in --tissue_classes

    Returns:
        List of tuples according to NewSegment input por tissues
    """
    tissues = []

    for i in range(1, 7):
        n_gaussians = 2

        if i == 4 or i == 5:
            n_gaussians = i - 1

        native_space = False
        dartel_input = False
        warped_unmodulated = False
        warped_modulated = False

        if i in tissue_classes:
            native_space = True
            if save_warped_unmodulated:
                warped_unmodulated = True
            if save_warped_modulated:
                warped_modulated = True

        if i in dartel_tissues:
            dartel_input = True

        tissues.append(
            (
                (tissue_map, i),
                n_gaussians,
                (native_space, dartel_input),
                (warped_unmodulated, warped_modulated),
            )
        )
    return tissues


class ApplySegmentationDeformationInput(SPMCommandInputSpec):

    deformation_field = File(
        exists=True,
        mandatory=True,
        field="comp{1}.def",
        desc="SPM Segmentation deformation file",
    )
    in_files = InputMultiPath(
        File(exists=True),
        mandatory=True,
        field="out{1}.pull.fnames",
        desc="Files on which deformation field is applied",
    )
    interpolation = traits.Range(
        low=0,
        high=7,
        field="out{1}.pull.interp",
        desc="degree of b-spline used for interpolation",
    )
    mask = traits.Int(
        0, usedefault=True, field="out{1}.pull.mask", desc="image masking"
    )
    fwhm = traits.List(
        traits.Float(0),
        field="out{1}.pull.fwhm",
        minlen=3,
        maxlen=3,
        desc="3-element list (opt)",
    )


class ApplySegmentationDeformationOutput(TraitedSpec):
    out_files = OutputMultiPath(File(exists=True), desc="Transformed files")


class ApplySegmentationDeformation(SPMCommand):
    """Uses SPM to apply a deformation field obtained from Segmentation routine to a given file

    Examples
    --------

    >>> import clinica.pipelines.t1_volume_tissue_segmentation.t1_volume_tissue_segmentation_utils as seg_utils
    >>> inv = seg_utils.ApplySegmentationDeformation()
    >>> inv.inputs.in_files = 'T1w.nii'
    >>> inv.inputs.deformation = 'y_T1w.nii'
    >>> inv.run() # doctest: +SKIP
    """

    input_spec = ApplySegmentationDeformationInput
    output_spec = ApplySegmentationDeformationOutput

    _jobtype = "util"
    _jobname = "defs"

    def _format_arg(self, opt, spec, val):
        """Convert input to appropriate format for spm"""
        if opt == "deformation_field":
            return np.array([list_to_filename(val)], dtype=object)
        if opt == "in_files":
            return np.array(filename_to_list(val), dtype=object)
        return val

    def _list_outputs(self):
        outputs = self._outputs().get()
        outputs["out_files"] = []
        for filename in self.inputs.in_files:
            _, fname = os.path.split(filename)
            outputs["out_files"].append(os.path.realpath("w%s" % fname))
        return outputs



# coding: utf8

import clinica.engine as ce


class T1FreeSurferVisualizer(ce.CmdParser):
    def define_name(self):
        """Define the sub-command name to run this pipeline."""
        self._name = "t1-freesurfer"

    def define_description(self):
        """Define a description of this pipeline."""
        self._description = (
            "Cross-sectional pre-processing of T1w images with FreeSurfer:\n"
            "https://aramislab.paris.inria.fr/clinica/docs/public/latest/Pipelines/T1_FreeSurfer/"
        )

    def define_options(self):
        """Define the sub-command arguments."""
        from clinica.engine.cmdparser import PIPELINE_CATEGORIES

        clinica_comp = self._args.add_argument_group(
            PIPELINE_CATEGORIES["CLINICA_COMPULSORY"]
        )
        clinica_comp.add_argument("caps_directory", help="Path to the CAPS directory.")
        clinica_comp.add_argument(
            "participant_id", help="Participant ID (e.g. sub-CLNC01)."
        )
        clinica_comp.add_argument("session_id", help="Session ID (e.g. ses-M00).")

    def run_command(self, args):
        """"""
        import os
        import subprocess

        from clinica.utils.check_dependency import check_freesurfer
        from clinica.utils.stream import cprint

        check_freesurfer()

        participant_id = args.participant_id
        session_id = args.session_id
        subject_id = participant_id + "_" + session_id

        cprint(
            "Visualizing outputs of t1-freesurfer pipeline for "
            + participant_id
            + " at session "
            + session_id
        )
        caps_participant = os.path.join(
            args.caps_directory,
            "subjects",
            participant_id,
            session_id,
            "t1",
            "freesurfer_cross_sectional",
            subject_id,
        )

        # Check that
        # 1) subject and session are present in CAPS
        # 2) a FreeSurfer folder has been launched
        if not os.path.exists(caps_participant):
            if os.path.exists(
                os.path.join(
                    args.caps_directory, "subjects", participant_id, session_id
                )
            ):
                raise RuntimeError(
                    f"Subject {participant_id} and session {session_id} exists "
                    f"in CAPS but no FreeSurfer related directory found"
                )
            else:
                raise RuntimeError(
                    f"Subject {participant_id} and session {session_id} are not "
                    f"present in CAPS {args.caps_directory}"
                )

        # Check that freesurfer has run properly
        log_file = os.path.join(caps_participant, "scripts", "recon-all-status.log")
        if os.path.isfile(log_file):
            last_line = str(subprocess.check_output(["tail", "-1", log_file]))
            if "finished without error" not in last_line.lower():
                raise ValueError(
                    f"FreeSurfer did not mark subject {subject_id} as -finished "
                    f"without error-. If the pipeline is still running, wait for it "
                    f"to finish. Otherwise, relaunch it."
                )
        else:
            raise FileNotFoundError(
                f"File {log_file} was not found. t1-freesurfer must be run "
                f"before using: clinica visualize t1-freesurfer"
            )

        # Check that files exists
        files = {
            "nu": os.path.join(caps_participant, "mri", "nu.mgz"),
            "aseg": os.path.join(caps_participant, "mri", "aseg.mgz"),
            "lh_white": os.path.join(caps_participant, "surf", "lh.white"),
            "lh_pial": os.path.join(caps_participant, "surf", "lh.pial"),
            "rh_white": os.path.join(caps_participant, "surf", "rh.white"),
            "rh_pial": os.path.join(caps_participant, "surf", "rh.pial"),
        }

        files_not_found = []
        for element in files:
            if not os.path.exists(files[element]):
                files_not_found.append(files[element])
        if len(files_not_found) > 0:
            error_string = "The following file(s) has(ve) not been found :\n"
            for f in files_not_found:
                error_string = error_string + f + "\n"
            raise FileNotFoundError(error_string)

        # Launch command
        command_line = (
            f"freeview"
            f" -v"
            f" {files['nu']}:name=T1w-nu-corrected"
            f" {files['aseg']}:colormap=lut:opacity=0.2:name=Segmentation"
            f" -f"
            f" {files['lh_white']}:edgecolor=blue:name=Left_WM/GM_Interface"
            f" {files['lh_pial']}:edgecolor=green:name=Left_GM/CSF_Interface"
            f" {files['rh_white']}:edgecolor=blue:name=Right_WM/GM_Interface"
            f" {files['rh_pial']}:edgecolor=green:name=Right_GM/CSF_Interface"
        )
        cprint(f"Running the following command: {command_line}")
        os.system(command_line)



import click

from .deeplearning_prepare_data import deeplearning_prepare_data_cli
from .dwi_connectome import dwi_connectome_cli
from .dwi_dti import dwi_dti_cli
from .dwi_preprocessing_using_phasediff_fieldmap import (
    dwi_preprocessing_using_phasediff_fieldmap_cli,
)
from .dwi_preprocessing_using_t1 import dwi_preprocessing_using_t1_cli
from .machine_learning import classification_cli
from .machine_learning_spatial_svm import spatial_svm_cli
from .pet_linear import pet_linear_cli
from .pet_surface import pet_surface_cli, pet_surface_longitudinal_cli
from .pet_volume import pet_volume_cli
from .statistics_surface import statistics_surface_cli
from .statistics_volume import statistics_volume_cli
from .statistics_volume_correction import statistics_volume_correction_cli
from .t1_freesurfer import t1_freesurfer_cli
from .t1_freesurfer_longitudinal import (
    t1_freesurfer_longitudinal_cli,
    t1_freesurfer_longitudinal_correction_cli,
    t1_freesurfer_template_cli,
)
from .t1_linear import t1_linear_cli
from .t1_volume import t1_volume_cli
from .t1_volume_create_dartel import t1_volume_create_dartel_cli
from .t1_volume_dartel2mni import t1_volume_dartel2mni_cli
from .t1_volume_existing_template import t1_volume_existing_template_cli
from .t1_volume_parcellation import t1_volume_parcellation_cli
from .t1_volume_register_dartel import t1_volume_register_dartel_cli
from .t1_volume_tissue_segmentation import t1_volume_tissue_segmentation_cli


class RegistrationOrderGroup(click.Group):
    """CLI group which lists commands by order or registration."""

    def list_commands(self, ctx):
        return self.commands.keys()


@click.group(cls=RegistrationOrderGroup, name="run")
def cli() -> None:
    """Run pipelines on BIDS and CAPS datasets."""
    pass


# Standard pipelines.
cli.add_command(t1_freesurfer_cli.cli)
cli.add_command(t1_volume_cli.cli)
cli.add_command(t1_freesurfer_longitudinal_cli.cli)
cli.add_command(t1_linear_cli.cli)
cli.add_command(dwi_preprocessing_using_phasediff_fieldmap_cli.cli)
cli.add_command(dwi_preprocessing_using_t1_cli.cli)
cli.add_command(dwi_dti_cli.cli)
cli.add_command(dwi_connectome_cli.cli)
cli.add_command(pet_linear_cli.cli)
cli.add_command(pet_volume_cli.cli)
cli.add_command(pet_surface_cli.cli)
cli.add_command(pet_surface_longitudinal_cli.cli)
cli.add_command(deeplearning_prepare_data_cli.cli)
cli.add_command(spatial_svm_cli.cli)
cli.add_command(classification_cli.cli)
cli.add_command(statistics_surface_cli.cli)
cli.add_command(statistics_volume_cli.cli)
cli.add_command(statistics_volume_correction_cli.cli)
cli.add_command(t1_volume_existing_template_cli.cli)
cli.add_command(t1_volume_tissue_segmentation_cli.cli)
cli.add_command(t1_volume_create_dartel_cli.cli)
cli.add_command(t1_volume_register_dartel_cli.cli)
cli.add_command(t1_volume_dartel2mni_cli.cli)
cli.add_command(t1_volume_parcellation_cli.cli)
cli.add_command(t1_freesurfer_template_cli.cli)
cli.add_command(t1_freesurfer_longitudinal_correction_cli.cli)


if __name__ == "__main__":
    cli()



# coding: utf8


def susceptibility_distortion_correction_using_phasediff_fmap(
        fugue_params=dict(smooth3d=2.0),
        register_fmap_on_b0=True,
        name='susceptibility_distortion_correction_using_phasediff_fmap'):
    """
    The fieldmap based method implements susceptibility distortion correction
    by using a mapping of the B0 field as proposed by [Jezzard95]_.
    This workflow uses the implementation of FSL
    (`FUGUE <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FUGUE>`_). Phase
    unwrapping is performed using
    `PRELUDE <http://fsl.fmrib.ox.ac.uk/fsl/fsl-4.1.9/fugue/prelude.html>`_
    [Jenkinson03]_. Preparation of the fieldmap is performed reproducing the
    script in FSL `fsl_prepare_fieldmap
    <http://fsl.fmrib.ox.ac.uk/fsl/fslwiki/FUGUE/Guide#SIEMENS_data>`_.

    Args:
        fugue_params: Regularisation to apply to the fieldmap with a
            3D Gaussian smoothing (default: 2.0)
        register_fmap_on_b0: Register the fmap on the b0 or not. If the fmap
            is already well aligned and
        name:

    Inputnode
        in_file (str): DWI dataset.
        in_bval (str): Bval file.
        in_mask (str): Mask file.
        in_fmap_magnitude (str): Magnitude fieldmap. Chose the fmap with the
            best contrast.
        in_fmap_phasediff (str): Phase difference fieldmap.
        delta_echo_time (float): Difference of echo time of the
        effective_echo_spacing (float):
        phase_encoding_direction (str):

    Outputnode:
        out_file (str): Output.
        out_vsm (str): The set of DWI volumes.
        out_warp (str): The bvalues corresponding to the out_dwi.


    .. warning:: Only SIEMENS format fieldmaps are supported.

    .. admonition:: References
<
      .. [Jezzard95] Jezzard P, and Balaban RS, `Correction for geometric
        distortion in echo planar images from B0 field variations
        <http://dx.doi.org/10.1002/mrm.1910340111>`_,
        MRM 34(1):65-73. (1995). doi: 10.1002/mrm.1910340111.

      .. [Jenkinson03] Jenkinson M., `Fast, automated, N-dimensional
        phase-unwrapping algorithm <http://dx.doi.org/10.1002/mrm.10354>`_,
        MRM 49(1):193-197, 2003, doi: 10.1002/mrm.10354.
    """
    import nipype.interfaces.ants as ants
    import nipype.interfaces.fsl as fsl
    import nipype.interfaces.utility as niu
    import nipype.pipeline.engine as pe
    from nipype.workflows.dmri.fsl.utils import (
        add_empty_vol,
        cleanup_edge_pipeline,
        demean_image,
        rads2radsec,
        siemens2rads,
        vsm2warp,
    )

    from clinica.utils.epi import bids_dir_to_fsl_dir
    from clinica.utils.fmap import resample_fmap_to_b0

    inputnode = pe.Node(niu.IdentityInterface(
        fields=['in_dwi', 'in_mask', 'in_fmap_phasediff', 'in_fmap_magnitude',
                'delta_echo_time', 'effective_echo_spacing',
                'phase_encoding_direction']),
        name='inputnode')
    outputnode = pe.Node(niu.IdentityInterface(
        fields=['out_file', 'out_vsm', 'out_warp', 'out_prepared_fmap']),
        name='outputnode')

    fsl_dir = pe.Node(niu.Function(
        input_names=['bids_dir'],
        output_names=['fsl_dir'],
        function=bids_dir_to_fsl_dir), name='ConvertToFslDir')

    get_b0 = pe.Node(fsl.ExtractROI(t_min=0, t_size=1), name='GetB0')

    first_mag = pe.Node(fsl.ExtractROI(t_min=0, t_size=1), name='GetFirst')

    n4 = pe.Node(ants.N4BiasFieldCorrection(dimension=3), name='N4MagnitudeFmap')

    bet = pe.Node(fsl.BET(frac=0.4, mask=True), name='BetN4MagnitudeFmap')

    dilate = pe.Node(
            fsl.maths.MathsCommand(
                nan2zeros=True,
                args='-kernel sphere 5 -dilM'),
            name='DilateBet')

    pha2rads = pe.Node(niu.Function(
        input_names=['in_file'],
        output_names=['out_file'],
        function=siemens2rads), name='PreparePhase')

    prelude = pe.Node(fsl.PRELUDE(process3d=True), name='PhaseUnwrap')

    rad2rsec = pe.Node(niu.Function(
        input_names=['in_file', 'delta_te'],
        output_names=['out_file'],
        function=rads2radsec), name='ToRadSec')

    # Node when we register the fmap onto the b0
    fmm2b0 = pe.Node(ants.Registration(
        output_warped_image=True), name="FmapMagnitudeToB0")
    fmm2b0.inputs.transforms = ['Rigid'] * 2
    fmm2b0.inputs.transform_parameters = [(1.0,)] * 2
    fmm2b0.inputs.number_of_iterations = [[50], [20]]
    fmm2b0.inputs.dimension = 3
    fmm2b0.inputs.metric = ['Mattes', 'Mattes']
    fmm2b0.inputs.metric_weight = [1.0] * 2
    fmm2b0.inputs.radius_or_number_of_bins = [64, 64]
    fmm2b0.inputs.sampling_strategy = ['Regular', 'Random']
    fmm2b0.inputs.sampling_percentage = [None, 0.2]
    fmm2b0.inputs.convergence_threshold = [1.e-5, 1.e-8]
    fmm2b0.inputs.convergence_window_size = [20, 10]
    fmm2b0.inputs.smoothing_sigmas = [[6.0], [2.0]]
    fmm2b0.inputs.sigma_units = ['vox'] * 2
    fmm2b0.inputs.shrink_factors = [[6], [1]]  # ,[1] ]
    fmm2b0.inputs.use_estimate_learning_rate_once = [True] * 2
    fmm2b0.inputs.use_histogram_matching = [True] * 2
    fmm2b0.inputs.initial_moving_transform_com = 0
    fmm2b0.inputs.collapse_output_transforms = True
    fmm2b0.inputs.winsorize_upper_quantile = 0.995

    # Node when we resample the fmap onto the b0
    res_fmap = pe.Node(niu.Function(
        input_names=['in_fmap', 'in_b0', 'out_file'],
        output_names=['out_resampled_fmap'],
        function=resample_fmap_to_b0), name='ResampleFmap')

    apply_xfm = pe.Node(ants.ApplyTransforms(
        dimension=3, interpolation='BSpline'), name='FmapPhaseToB0')

    pre_fugue = pe.Node(fsl.FUGUE(save_fmap=True), name='PreliminaryFugue')

    demean = pe.Node(
            niu.Function(
                input_names=['in_file', 'in_mask'],
                output_names=['out_file'],
                function=demean_image),
            name='DemeanFmap')

    cleanup = cleanup_edge_pipeline()

    add_vol = pe.Node(niu.Function(
        input_names=['in_file'],
        output_names=['out_file'],
        function=add_empty_vol), name='AddEmptyVol')

    vsm = pe.Node(fsl.FUGUE(save_shift=True, **fugue_params),
                  name="ComputeVSM")

    split = pe.Node(fsl.Split(dimension='t'), name='SplitDWIs')

    merge = pe.Node(fsl.Merge(dimension='t'), name='MergeDWIs')

    unwarp = pe.MapNode(fsl.FUGUE(icorr=True, forward_warping=False),
                        iterfield=['in_file'], name='UnwarpDWIs')

    thres = pe.MapNode(fsl.Threshold(thresh=0.0), iterfield=['in_file'],
                       name='RemoveNegative')

    vsm2dfm = vsm2warp()
    vsm2dfm.inputs.inputnode.scaling = 1.0

    wf = pe.Workflow(name=name)
    wf.connect([
        (inputnode, fsl_dir, [('phase_encoding_direction', 'bids_dir')]),  # noqa
        (inputnode, pha2rads, [('in_fmap_phasediff', 'in_file')]),  # noqa
        (inputnode, get_b0,   [('in_dwi', 'in_file')]),  # noqa
        (inputnode, first_mag, [('in_fmap_magnitude', 'in_file')]),  # noqa
        (first_mag, n4, [('roi_file', 'input_image')]),  # noqa
        (n4, bet, [('output_image', 'in_file')]),  # noqa
        (bet, dilate, [('mask_file', 'in_file')]),  # noqa
        (pha2rads, prelude, [('out_file',     'phase_file')]),  # noqa
        (n4,       prelude, [('output_image', 'magnitude_file')]),  # noqa
        (dilate,   prelude, [('out_file',     'mask_file')]),  # noqa
        (prelude,   rad2rsec, [('unwrapped_phase_file', 'in_file')]),  # noqa
        (inputnode, rad2rsec, [('delta_echo_time',      'delta_te')]),  # noqa
    ])
    if register_fmap_on_b0:
        wf.connect([
            (get_b0,    fmm2b0, [('roi_file',     'fixed_image')]),  # noqa
            (n4,        fmm2b0, [('output_image', 'moving_image')]),  # noqa
            (inputnode, fmm2b0, [('in_mask',      'fixed_image_mask')]),  # noqa
            (dilate,    fmm2b0, [('out_file',     'moving_image_mask')]),  # noqa
            (get_b0,   apply_xfm, [('roi_file',            'reference_image')]),  # noqa
            (rad2rsec, apply_xfm, [('out_file',            'input_image')]),  # noqa
            (fmm2b0,   apply_xfm, [('forward_transforms',  'transforms'),  # noqa
                                   ('forward_invert_flags', 'invert_transform_flags')]),  # noqa
            (apply_xfm, pre_fugue, [('output_image', 'fmap_in_file')]),  # noqa
            (inputnode, pre_fugue, [('in_mask', 'mask_file')]),  # noqa

            (apply_xfm, outputnode, [('output_image', 'out_prepared_fmap')])  # noqa
        ])
    else:
        wf.connect([
            (get_b0,   res_fmap, [('roi_file', 'in_b0')]),  # noqa
            (rad2rsec, res_fmap, [('out_file', 'in_fmap')]),  # noqa
            (res_fmap,  pre_fugue, [('out_resampled_fmap', 'fmap_in_file')]),  # noqa
            (inputnode, pre_fugue, [('in_mask',            'mask_file')]),  # noqa

            (res_fmap, outputnode, [('out_resampled_fmap', 'out_prepared_fmap')])  # noqa
        ])
    wf.connect([
        (pre_fugue, demean, [('fmap_out_file', 'in_file')]),  # noqa
        (inputnode, demean, [('in_mask', 'in_mask')]),  # noqa
        (demean,    cleanup, [('out_file', 'inputnode.in_file')]),  # noqa
        (inputnode, cleanup, [('in_mask',  'inputnode.in_mask')]),  # noqa
        (cleanup, add_vol, [('outputnode.out_file', 'in_file')]),  # noqa
        (inputnode, vsm, [('in_mask',                'mask_file')]),  # noqa
        (add_vol,   vsm, [('out_file',               'fmap_in_file')]),  # noqa
        (inputnode, vsm, [('delta_echo_time',        'asym_se_time')]),  # noqa
        (inputnode, vsm, [('effective_echo_spacing', 'dwell_time')]),  # noqa
        (inputnode, split, [('in_dwi', 'in_file')]),  # noqa
        (split,   unwarp, [('out_files',      'in_file')]),  # noqa
        (vsm,     unwarp, [('shift_out_file', 'shift_in_file')]),  # noqa
        (fsl_dir, unwarp, [('fsl_dir',        'unwarp_direction')]),  # noqa
        (unwarp, thres, [('unwarped_file', 'in_file')]),  # noqa
        (thres, merge, [('out_file', 'in_files')]),  # noqa
        (merge,   vsm2dfm, [('merged_file',    'inputnode.in_ref')]),  # noqa
        (vsm,     vsm2dfm, [('shift_out_file', 'inputnode.in_vsm')]),  # noqa
        (fsl_dir, vsm2dfm, [('fsl_dir',        'inputnode.enc_dir')]),  # noqa
        (rad2rsec, outputnode, [('out_file',            'out_native_fmap')]),  # noqa
        (merge,    outputnode, [('merged_file',         'out_file')]),  # noqa
        (vsm,      outputnode, [('shift_out_file',      'out_vsm')]),  # noqa
        (vsm2dfm,  outputnode, [('outputnode.out_warp', 'out_warp')]),  # noqa
    ])

    return wf



# coding: utf8

import clinica.pipelines.engine as cpe

# Use hash instead of parameters for iterables folder names
from nipype import config

cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class DwiConnectome(cpe.Pipeline):
    """Connectome-based processing of corrected DWI datasets.

    Returns:
        A clinica pipeline object containing the DwiConnectome pipeline.
    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        if self.parameters["n_tracks"] < 0:
            raise ValueError(
                f"The n_tracks parameter ({self.parameters['n_tracks']}) should be positive."
            )

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """

        return [
            "t1_brain_file",
            "wm_mask_file",
            "dwi_file",
            "dwi_brainmask_file",
            "grad_fsl",
            "atlas_files",
        ]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """

        return ["response", "fod", "tracts", "nodes", "connectomes"]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import re

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.utils.input_files as input_files
        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint

        all_errors = []

        # Inputs from t1-freesurfer pipeline
        # ==================================

        # White matter segmentation
        try:
            wm_mask_files = clinica_file_reader(
                self.subjects, self.sessions, self.caps_directory, input_files.T1_FS_WM
            )
        except ClinicaException as e:
            all_errors.append(e)
        # Desikan parcellation
        try:
            aparc_aseg_files = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_DESIKAN,
            )
        except ClinicaException as e:
            all_errors.append(e)
        # Destrieux parcellation
        try:
            aparc_aseg_a2009s_files = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_DESTRIEUX,
            )
        except ClinicaException as e:
            all_errors.append(e)
        # Inputs from dwi-preprocessing pipeline
        # ======================================
        # Preprocessed DWI
        try:
            dwi_files = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.DWI_PREPROC_NII,
            )
        except ClinicaException as e:
            all_errors.append(e)
        # B0 brainmask
        try:
            dwi_brainmask_files = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.DWI_PREPROC_BRAINMASK,
            )
        except ClinicaException as e:
            all_errors.append(e)
        # Preprocessed bvec
        try:
            bvec_files = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.DWI_PREPROC_BVEC,
            )
        except ClinicaException as e:
            all_errors.append(e)
        # Preprocessed bval
        try:
            bval_files = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.DWI_PREPROC_BVAL,
            )
        except ClinicaException as e:
            all_errors.append(e)
        if len(all_errors) > 0:
            error_message = "Clinica faced errors while trying to read files in your BIDS or CAPS directories.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaCAPSError(error_message)
        # Check space of DWI dataset
        dwi_file_spaces = [
            re.search(".*_space-(.*)_preproc.nii.*", file, re.IGNORECASE).group(1)
            for file in dwi_files
        ]

        # Return an error if all the DWI files are not in the same space
        if any(a != dwi_file_spaces[0] for a in dwi_file_spaces):
            raise ClinicaCAPSError(
                "Preprocessed DWI files are not all in the same space. "
                "Please process them separately using the appropriate subjects/sessions `.tsv` file (-tsv option)."
            )
        # Used only for for T1-B0 registration
        if dwi_file_spaces[0] == "b0":
            # Brain extracted T1w
            t1_brain_files = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_BRAIN,
            )
        list_atlas_files = [
            [aparc_aseg, aparc_aseg_a2009]
            for aparc_aseg, aparc_aseg_a2009 in zip(
                aparc_aseg_files, aparc_aseg_a2009s_files
            )
        ]

        list_grad_fsl = [(bvec, bval) for bvec, bval in zip(bvec_files, bval_files)]

        p_id_images_to_process = [
            re.search(r"(sub-[a-zA-Z0-9]+)", caps_file).group()
            for caps_file in dwi_files
        ]
        s_id_images_to_process = [
            re.search(r"(ses-[a-zA-Z0-9]+)", caps_file).group()
            for caps_file in dwi_files
        ]
        images_to_process = ", ".join(
            p_id[4:] + "|" + s_id[4:]
            for p_id, s_id in zip(p_id_images_to_process, s_id_images_to_process)
        )
        cprint(
            f"The pipeline will be run on the following subject(s): {images_to_process}"
        )

        if dwi_file_spaces[0] == "b0":
            self.parameters["dwi_space"] = "b0"
            read_node = npe.Node(
                name="ReadingFiles",
                iterables=[
                    ("wm_mask_file", wm_mask_files),
                    ("t1_brain_file", t1_brain_files),
                    ("dwi_file", dwi_files),
                    ("dwi_brainmask_file", dwi_brainmask_files),
                    ("grad_fsl", list_grad_fsl),
                    ("atlas_files", list_atlas_files),
                ],
                synchronize=True,
                interface=nutil.IdentityInterface(fields=self.get_input_fields()),
            )
            # fmt: off
            self.connect(
                [
                    (read_node, self.input_node, [("t1_brain_file", "t1_brain_file")]),
                    (read_node, self.input_node, [("wm_mask_file", "wm_mask_file")]),
                    (read_node, self.input_node, [("dwi_file", "dwi_file")]),
                    (read_node, self.input_node, [("dwi_brainmask_file", "dwi_brainmask_file")]),
                    (read_node, self.input_node, [("grad_fsl", "grad_fsl")]),
                    (read_node, self.input_node, [("atlas_files", "atlas_files")]),
                ]
            )
            # fmt: on
        elif dwi_file_spaces[0] == "T1w":
            self.parameters["dwi_space"] = "T1w"
            read_node = npe.Node(
                name="ReadingFiles",
                iterables=[
                    ("wm_mask_file", wm_mask_files),
                    ("dwi_file", dwi_files),
                    ("dwi_brainmask_file", dwi_brainmask_files),
                    ("grad_fsl", list_grad_fsl),
                    ("atlas_files", list_atlas_files),
                ],
                synchronize=True,
                interface=nutil.IdentityInterface(fields=self.get_input_fields()),
            )
            # fmt: off
            self.connect(
                [
                    (read_node, self.input_node, [("wm_mask_file", "wm_mask_file")]),
                    (read_node, self.input_node, [("dwi_file", "dwi_file")]),
                    (read_node, self.input_node, [("dwi_brainmask_file", "dwi_brainmask_file")]),
                    (read_node, self.input_node, [("grad_fsl", "grad_fsl")]),
                    (read_node, self.input_node, [("atlas_files", "atlas_files")]),
                ]
            )
            # fmt: on
        else:
            raise ClinicaCAPSError(
                "Bad preprocessed DWI space. Please check your CAPS folder."
            )

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.io as nio
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.pipelines.dwi_connectome.dwi_connectome_utils as utils

        # Writing CAPS
        # ============
        join_node = npe.JoinNode(
            name="JoinOutputs",
            joinsource="ReadingFiles",
            interface=nutil.IdentityInterface(fields=self.get_output_fields()),
        )

        write_node = npe.MapNode(
            name="WritingCAPS",
            iterfield=["container"]
            + [f"connectome_based_processing.@{o}" for o in self.get_output_fields()],
            interface=nio.DataSink(
                infields=[
                    f"connectome_based_processing.@{o}"
                    for o in self.get_output_fields()
                ]
            ),
        )
        write_node.inputs.base_directory = self.caps_directory
        write_node.inputs.container = utils.get_containers(self.subjects, self.sessions)
        write_node.inputs.substitutions = [("trait_added", "")]
        write_node.inputs.parameterization = False

        # fmt: off
        self.connect(
            [
                # Writing CAPS
                (self.output_node, join_node, [("response", "response")]),
                (self.output_node, join_node, [("fod", "fod")]),
                (self.output_node, join_node, [("tracts", "tracts")]),
                (self.output_node, join_node, [("nodes", "nodes")]),
                (self.output_node, join_node, [("connectomes", "connectomes")]),
                (join_node, write_node, [("response", "connectome_based_processing.@response")]),
                (join_node, write_node, [("fod", "connectome_based_processing.@fod")]),
                (join_node, write_node, [("tracts", "connectome_based_processing.@tracts")]),
                (join_node, write_node, [("nodes", "connectome_based_processing.@nodes")]),
                (join_node, write_node, [("connectomes", "connectome_based_processing.@connectomes")]),
            ]
        )
        # fmt: on

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline.

        Notes:
            - If `FSLOUTPUTTYPE` environment variable is not set, `nipype` takes
            NIFTI by default.

        Todo:
            - [x] Detect space automatically.
            - [ ] Allow for custom parcellations (See TODOs in utils).

        """
        import nipype.interfaces.freesurfer as fs
        import nipype.interfaces.fsl as fsl
        import nipype.interfaces.mrtrix3 as mrtrix3
        import nipype.interfaces.utility as niu
        import nipype.pipeline.engine as npe

        import clinica.pipelines.dwi_connectome.dwi_connectome_utils as utils
        from clinica.lib.nipype.interfaces.mrtrix3.reconst import EstimateFOD
        from clinica.lib.nipype.interfaces.mrtrix3.tracking import Tractography
        from clinica.lib.nipype.interfaces.mrtrix.preprocess import MRTransform
        from clinica.utils.exceptions import ClinicaCAPSError, ClinicaException
        from clinica.utils.mri_registration import (
            convert_flirt_transformation_to_mrtrix_transformation,
        )
        from clinica.utils.stream import cprint

        # Nodes
        # =====
        # B0 Extraction (only if space=b0)
        # -------------
        split_node = npe.Node(name="Reg-0-DWI-B0Extraction", interface=fsl.Split())
        split_node.inputs.output_type = "NIFTI_GZ"
        split_node.inputs.dimension = "t"
        select_node = npe.Node(name="Reg-0-DWI-B0Selection", interface=niu.Select())
        select_node.inputs.index = 0

        # B0 Brain Extraction (only if space=b0)
        # -------------------
        mask_node = npe.Node(name="Reg-0-DWI-BrainMasking", interface=fsl.ApplyMask())
        mask_node.inputs.output_type = "NIFTI_GZ"

        # T1-to-B0 Registration (only if space=b0)
        # ---------------------
        t12b0_reg_node = npe.Node(
            name="Reg-1-T12B0Registration",
            interface=fsl.FLIRT(
                dof=6,
                interp="spline",
                cost="normmi",
                cost_func="normmi",
            ),
        )
        t12b0_reg_node.inputs.output_type = "NIFTI_GZ"

        # MGZ File Conversion (only if space=b0)
        # -------------------
        t1_brain_conv_node = npe.Node(
            name="Reg-0-T1-T1BrainConvertion", interface=fs.MRIConvert()
        )
        wm_mask_conv_node = npe.Node(
            name="Reg-0-T1-WMMaskConvertion", interface=fs.MRIConvert()
        )

        # WM Transformation (only if space=b0)
        # -----------------
        wm_transform_node = npe.Node(
            name="Reg-2-WMTransformation", interface=fsl.ApplyXFM()
        )
        wm_transform_node.inputs.apply_xfm = True

        # Nodes Generation
        # ----------------
        label_convert_node = npe.MapNode(
            name="0-LabelsConversion",
            iterfield=["in_file", "in_config", "in_lut", "out_file"],
            interface=mrtrix3.LabelConvert(),
        )
        label_convert_node.inputs.in_config = utils.get_conversion_luts()
        label_convert_node.inputs.in_lut = utils.get_luts()

        # FSL flirt matrix to MRtrix matrix Conversion (only if space=b0)
        # --------------------------------------------
        fsl2mrtrix_conv_node = npe.Node(
            name="Reg-2-FSL2MrtrixConversion",
            interface=niu.Function(
                input_names=[
                    "in_source_image",
                    "in_reference_image",
                    "in_flirt_matrix",
                    "name_output_matrix",
                ],
                output_names=["out_mrtrix_matrix"],
                function=convert_flirt_transformation_to_mrtrix_transformation,
            ),
        )

        # Parc. Transformation (only if space=b0)
        # --------------------
        parc_transform_node = npe.MapNode(
            name="Reg-2-ParcTransformation",
            iterfield=["in_files", "out_filename"],
            interface=MRTransform(),
        )

        # Response Estimation
        # -------------------
        resp_estim_node = npe.Node(
            name="1a-ResponseEstimation", interface=mrtrix3.ResponseSD()
        )
        resp_estim_node.inputs.algorithm = "tournier"

        # FOD Estimation
        # --------------
        fod_estim_node = npe.Node(name="1b-FODEstimation", interface=EstimateFOD())
        fod_estim_node.inputs.algorithm = "csd"

        # Tracts Generation
        # -----------------
        tck_gen_node = npe.Node(name="2-TractsGeneration", interface=Tractography())
        tck_gen_node.inputs.select = self.parameters["n_tracks"]
        tck_gen_node.inputs.algorithm = "iFOD2"

        # Connectome Generation
        # ---------------------
        # only the parcellation and output filename should be iterable, the tck
        # file stays the same.
        conn_gen_node = npe.MapNode(
            name="3-ConnectomeGeneration",
            iterfield=["in_parc", "out_file"],
            interface=mrtrix3.BuildConnectome(),
        )

        # Print begin message
        # -------------------
        print_begin_message = npe.MapNode(
            interface=niu.Function(
                input_names=["in_bids_or_caps_file"],
                function=utils.print_begin_pipeline,
            ),
            iterfield="in_bids_or_caps_file",
            name="WriteBeginMessage",
        )

        # Print end message
        # -----------------
        print_end_message = npe.MapNode(
            interface=niu.Function(
                input_names=["in_bids_or_caps_file", "final_file"],
                function=utils.print_end_pipeline,
            ),
            iterfield=["in_bids_or_caps_file"],
            name="WriteEndMessage",
        )

        # CAPS File names Generation
        # --------------------------
        caps_filenames_node = npe.Node(
            name="CAPSFilenamesGeneration",
            interface=niu.Function(
                input_names="dwi_file",
                output_names=self.get_output_fields(),
                function=utils.get_caps_filenames,
            ),
        )

        # Connections
        # ===========
        # Computation of the diffusion model, tractography & connectome
        # -------------------------------------------------------------
        # fmt: off
        self.connect(
            [
                (self.input_node, print_begin_message, [("dwi_file", "in_bids_or_caps_file")]),
                (self.input_node, caps_filenames_node, [("dwi_file", "dwi_file")]),
                # Response Estimation
                (self.input_node, resp_estim_node, [("dwi_file", "in_file")]),  # Preproc. DWI
                (self.input_node, resp_estim_node, [("dwi_brainmask_file", "in_mask")]),  # B0 brain mask
                (self.input_node, resp_estim_node, [("grad_fsl", "grad_fsl")]),  # bvecs and bvals
                (caps_filenames_node, resp_estim_node, [("response", "wm_file")]),  # output response filename
                # FOD Estimation
                (self.input_node, fod_estim_node, [("dwi_file", "in_file")]),  # Preproc. DWI
                (resp_estim_node, fod_estim_node, [("wm_file", "wm_txt")]),  # Response (txt file)
                (self.input_node, fod_estim_node, [("dwi_brainmask_file", "mask_file")]),  # B0 brain mask
                (self.input_node, fod_estim_node, [("grad_fsl", "grad_fsl")]),  # T1-to-B0 matrix file
                (caps_filenames_node, fod_estim_node, [("fod", "wm_odf")]),  # output odf filename
                # Tracts Generation
                (fod_estim_node, tck_gen_node, [("wm_odf", "in_file")]),  # ODF file
                (caps_filenames_node, tck_gen_node, [("tracts", "out_file")]),  # output tck filename
                # Label Conversion
                (self.input_node, label_convert_node, [("atlas_files", "in_file")]),  # atlas image files
                (caps_filenames_node, label_convert_node, [("nodes", "out_file")]),  # converted atlas image filenames
                # Connectomes Generation
                (tck_gen_node, conn_gen_node, [("out_file", "in_file")]),
                (caps_filenames_node, conn_gen_node, [("connectomes", "out_file")]),
            ]
        )
        # Registration T1-DWI (only if space=b0)
        # -------------------
        if self.parameters["dwi_space"] == "b0":
            self.connect(
                [
                    # MGZ Files Conversion
                    (self.input_node, t1_brain_conv_node, [("t1_brain_file", "in_file")]),
                    (self.input_node, wm_mask_conv_node, [("wm_mask_file", "in_file")]),
                    # B0 Extraction
                    (self.input_node, split_node, [("dwi_file", "in_file")]),
                    (split_node, select_node, [("out_files", "inlist")]),
                    # Masking
                    (select_node, mask_node, [("out", "in_file")]),  # B0
                    (self.input_node, mask_node, [("dwi_brainmask_file", "mask_file")]),  # Brain mask
                    # T1-to-B0 Registration
                    (t1_brain_conv_node, t12b0_reg_node, [("out_file", "in_file")]),  # Brain
                    (mask_node, t12b0_reg_node, [("out_file", "reference")]),  # B0 brain-masked
                    # WM Transformation
                    (wm_mask_conv_node, wm_transform_node, [("out_file", "in_file")]),  # Brain mask
                    (mask_node, wm_transform_node, [("out_file", "reference")]),  # BO brain-masked
                    (t12b0_reg_node, wm_transform_node, [("out_matrix_file", "in_matrix_file")]),  # T1-to-B0 matrix file
                    # FSL flirt matrix to MRtrix matrix Conversion
                    (t1_brain_conv_node, fsl2mrtrix_conv_node, [("out_file", "in_source_image")]),
                    (mask_node, fsl2mrtrix_conv_node, [("out_file", "in_reference_image")]),
                    (t12b0_reg_node, fsl2mrtrix_conv_node, [("out_matrix_file", "in_flirt_matrix")]),
                    # Apply registration without resampling on parcellations
                    (label_convert_node, parc_transform_node, [("out_file", "in_files")]),
                    (fsl2mrtrix_conv_node, parc_transform_node, [("out_mrtrix_matrix", "linear_transform")]),
                    (caps_filenames_node, parc_transform_node, [("nodes", "out_filename")]),
                ]
            )
        # Special care for Parcellation & WM mask
        # ---------------------------------------
        if self.parameters["dwi_space"] == "b0":
            self.connect(
                [
                    (wm_transform_node, tck_gen_node, [("out_file", "seed_image")]),
                    (parc_transform_node, conn_gen_node, [("out_file", "in_parc")]),
                    (parc_transform_node, self.output_node, [("out_file", "nodes")]),
                ]
            )
        elif self.parameters["dwi_space"] == "T1w":
            self.connect(
                [
                    (self.input_node, tck_gen_node, [("wm_mask_file", "seed_image")]),
                    (label_convert_node, conn_gen_node, [("out_file", "in_parc")]),
                    (label_convert_node, self.output_node, [("out_file", "nodes")]),
                ]
            )
        else:
            raise ClinicaCAPSError(
                "Bad preprocessed DWI space. Please check your CAPS folder."
            )
        # Outputs
        # -------
        self.connect(
            [
                (resp_estim_node, self.output_node, [("wm_file", "response")]),
                (fod_estim_node, self.output_node, [("wm_odf", "fod")]),
                (tck_gen_node, self.output_node, [("out_file", "tracts")]),
                (conn_gen_node, self.output_node, [("out_file", "connectomes")]),
                (self.input_node, print_end_message, [("dwi_file", "in_bids_or_caps_file")]),
                (conn_gen_node, print_end_message, [("out_file", "final_file")]),
            ]
        )
        # fmt: off



"""Script that generate the skeleton for a new Pipeline."""
from os import getcwd

import click


@click.group("generate")
def cli():
    """Instantiate a new pipeline from available templates."""
    pass


@cli.command("template")
@click.argument("name")
@click.option(
    "-d",
    "--output-dir",
    type=click.Path(exists=True, writable=True),
    default=getcwd(),
    help="Target path to the generated template.",
)
def generate_template(name: str, output_dir: str):
    """Generate the skeleton for a new pipeline."""
    from os import mkdir
    from os.path import isdir, join, splitext

    from jinja2 import Environment, PackageLoader, select_autoescape

    pipeline = dict(
        title=name,
        module_name=name.replace(" ", "_").lower(),
        class_name=name.replace(" ", "_"),
        command_name=name.replace(" ", "-").lower(),
        dir=join(output_dir, name.replace(" ", "_").lower()),
    )

    env = Environment(
        loader=PackageLoader("clinica", "resources/templates/pipeline_template"),
        autoescape=select_autoescape(["py"]),
    )

    if not isdir(pipeline["dir"]):
        mkdir(pipeline["dir"])

    for template_file in env.list_templates(extensions="j2"):
        template = env.get_template(template_file)
        rendered_file, template_ext = splitext(template_file)
        rendered_filename, rendered_file_ext = splitext(rendered_file)
        if rendered_file_ext == ".py" and rendered_filename != "__init__":
            path_to_write = join(
                pipeline["dir"], pipeline["module_name"] + "_" + rendered_file
            )
        else:
            path_to_write = join(pipeline["dir"], rendered_file)
        with open(path_to_write, "w+") as file_to_write:
            print(f"Generating template {file_to_write.name}")
            file_to_write.write(template.render(pipeline=pipeline))


if __name__ == "__main__":
    cli()



# coding: utf-8

"""Module for converting T1 of ADNI."""


def convert_adni_t1(
    source_dir, csv_dir, dest_dir, conversion_dir, subjs_list=None, mod_to_update=False
):
    """Convert T1 MR images of ADNI into BIDS format.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        conversion_dir: path to the TSV files including the paths to original images
        subjs_list: subjects list
        mod_to_update: If True, pre-existing images in the BIDS directory will be erased and extracted again.
    """
    from os import path

    from pandas.io import parsers

    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids
    from clinica.utils.stream import cprint

    if not subjs_list:
        adni_merge_path = path.join(csv_dir, "ADNIMERGE.csv")
        adni_merge = parsers.read_csv(adni_merge_path, sep=",", low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint(
        f"Calculating paths of T1 images. Output will be stored in {conversion_dir}."
    )
    images = compute_t1_paths(source_dir, csv_dir, dest_dir, subjs_list, conversion_dir)
    cprint("Paths of T1 images found. Exporting images into BIDS ...")
    paths_to_bids(images, dest_dir, "t1", mod_to_update=mod_to_update)
    cprint(msg="T1 conversion done.", lvl="debug")


def compute_t1_paths(source_dir, csv_dir, dest_dir, subjs_list, conversion_dir):
    """Compute the paths to T1 MR images and store them in a TSV file.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list
        conversion_dir: path to the TSV files including the paths to original images

    Returns:
        images: a dataframe with all the paths to the T1 MR images that will be converted into BIDS
    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        find_image_path,
        visits_to_timepoints,
    )
    from clinica.utils.stream import cprint

    t1_col_df = [
        "Subject_ID",
        "VISCODE",
        "Visit",
        "Sequence",
        "Scan_Date",
        "Study_ID",
        "Field_Strength",
        "Series_ID",
        "Image_ID",
        "Original",
    ]
    t1_df = pd.DataFrame(columns=t1_col_df)
    t1_dfs_list = []

    # Loading needed .csv files
    adni_merge = pd.read_csv(
        path.join(csv_dir, "ADNIMERGE.csv"), sep=",", low_memory=False
    )
    mprage_meta = pd.read_csv(
        path.join(csv_dir, "MPRAGEMETA.csv"), sep=",", low_memory=False
    )
    mri_quality = pd.read_csv(
        path.join(csv_dir, "MRIQUALITY.csv"), sep=",", low_memory=False
    )
    mayo_mri_qc = pd.read_csv(
        path.join(csv_dir, "MAYOADIRL_MRI_IMAGEQC_12_08_15.csv"),
        sep=",",
        low_memory=False,
    )
    # Keep only T1 scans
    mayo_mri_qc = mayo_mri_qc[mayo_mri_qc.series_type == "T1"]

    # We will convert the images for each subject in the subject list
    for subj in subjs_list:

        # Filter ADNIMERGE, MPRAGE METADATA and QC for only one subject and sort the rows/visits by examination date
        adnimerge_subj = adni_merge[adni_merge.PTID == subj]
        adnimerge_subj = adnimerge_subj.sort_values("EXAMDATE")

        mprage_meta_subj = mprage_meta[mprage_meta.SubjectID == subj]
        mprage_meta_subj = mprage_meta_subj.sort_values("ScanDate")

        mri_quality_subj = mri_quality[mri_quality.RID == int(subj[-4:])]
        mayo_mri_qc_subj = mayo_mri_qc[mayo_mri_qc.RID == int(subj[-4:])]

        # Obtain corresponding timepoints for the subject visits
        visits = visits_to_timepoints(
            subj, mprage_meta_subj, adnimerge_subj, "T1", "Visit", "ScanDate"
        )

        for visit_info in visits.keys():
            cohort = visit_info[1]
            timepoint = visit_info[0]
            visit_str = visits[visit_info]

            image_dict = None

            if cohort in ("ADNI1", "ADNIGO", "ADNI2"):
                image_dict = adni1go2_image(
                    subj,
                    timepoint,
                    visit_str,
                    mprage_meta_subj,
                    mri_quality_subj,
                    mayo_mri_qc_subj,
                    preferred_field_strength=1.5 if cohort == "ADNI1" else 3.0,
                )
            elif cohort == "ADNI3":
                image_dict = adni3_image(
                    subj, timepoint, visit_str, mprage_meta_subj, mayo_mri_qc_subj
                )
            else:
                cprint(
                    f"Subject {subj} visit {visit_str} belongs to an unknown cohort: {cohort}"
                )

            if image_dict is not None:
                row_to_append = pd.DataFrame(
                    image_dict,
                    index=[
                        "i",
                    ],
                )
                t1_dfs_list.append(row_to_append)

    if t1_dfs_list:
        t1_df = pd.concat(t1_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # Eq_1
        ("031_S_0830", "m48"),
        ("100_S_0995", "m18"),
        ("031_S_0867", "m48"),
        ("100_S_0892", "m18"),
        ("003_S_6264", "m12"),
        # Empty folders
        # ('029_S_0845', 'm24'),
        # ('094_S_1267', 'm24'),
        # ('029_S_0843', 'm24'),
        # ('027_S_0307', 'm48'),
        # ('057_S_1269', 'm24'),
        # ('036_S_4899', 'm03'),
        # ('033_S_1016', 'm120'),
        # ('130_S_4984', 'm12'),
        # ('027_S_4802', 'm06'),
        # ('131_S_0409', 'bl'),
        # ('082_S_4224', 'm24'),
        # ('006_S_4960', 'bl'),
        # ('006_S_4960', 'm03'),
        # ('006_S_4960', 'm06'),
        # ('006_S_4960', 'm12'),
        # ('006_S_4960', 'm24'),
        # ('006_S_4960', 'm36'),
        # ('006_S_4960', 'm72'),
        # ('022_S_5004', 'bl'),
        # ('022_S_5004', 'm03'),
        # Several images: T1wa ...
        ("006_S_4485", "m84"),
        ("029_S_2395", "m72"),
        ("114_S_6039", "bl"),
        ("016_S_4952", "m48"),
    ]

    # Removing known exceptions from images to convert
    if not t1_df.empty:
        error_indices = t1_df.index[
            t1_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors), axis=1
            )
        ]
        t1_df.drop(error_indices, inplace=True)

    # Checking for images paths in filesystem
    images = find_image_path(t1_df, source_dir, "T1", "S", "Series_ID")
    images.to_csv(path.join(conversion_dir, "t1_paths.tsv"), sep="\t", index=False)

    return images


def adni1go2_image(
    subject_id,
    timepoint,
    visit_str,
    mprage_meta_subj,
    mri_quality_subj,
    mayo_mri_qc_subj,
    preferred_field_strength=3.0,
):
    """Select the preferred scan for a subject in a visit, given the subject belongs to ADNI 1, Go or 2 cohorts.

    Args:

        subject_id: string containing subject ID
        timepoint: string of visit code in months
        visit_str: string of visit name
        mprage_meta_subj: DataFrame of MPRAGE metadata of images corresponding to the subject
        mri_quality_subj: DatFrame of MR image quality of images corresponding to the subject
        mayo_mri_qc_subj: DatFrame of MAYO Clinic MR image quality of images corresponding to the subject
        preferred_field_strength: Field strength that is preferred in case there are several image acquisitions

    Returns:
        Dictionary containing selected scan information
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        replace_sequence_chars,
    )

    # filter out images that do not pass QC
    mprage_meta_subj = mprage_meta_subj[
        mprage_meta_subj.apply(
            lambda x: check_qc(x, subject_id, visit_str, mri_quality_subj), axis=1
        )
    ]
    filtered_mprage = preferred_processed_scan(mprage_meta_subj, visit_str)

    # If no N3 processed image found (it means there are no processed images at all), get best original image
    if filtered_mprage.empty:
        return original_image(
            subject_id,
            timepoint,
            visit_str,
            mprage_meta_subj,
            mayo_mri_qc_subj,
            preferred_field_strength,
        )

    # If there are images with different magnetic field strength, prefer 1.5T images for ADNI1, 3.0T otherwise
    if len(filtered_mprage.MagStrength.unique()) > 1:
        filtered_mprage = filtered_mprage[
            filtered_mprage.MagStrength == preferred_field_strength
        ]

    # Sort by Series ID in case there are several images, so we keep the one acquired first
    filtered_mprage = filtered_mprage.sort_values("SeriesID")
    scan = filtered_mprage.iloc[0]

    n3 = scan.Sequence.find("N3")
    # Sequence ends in 'N3' or in 'N3m'
    sequence = scan.Sequence[: n3 + 2 + int(scan.Sequence[n3 + 2] == "m")]
    sequence = replace_sequence_chars(sequence)

    return {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": sequence,
        "Scan_Date": scan.ScanDate,
        "Study_ID": str(scan.StudyID),
        "Series_ID": str(scan.SeriesID),
        "Image_ID": str(scan.ImageUID),
        "Field_Strength": scan.MagStrength,
        "Original": False,
    }


def preferred_processed_scan(mprage_meta_subj, visit_str, unwanted_series_id=()):
    """Filter content of Dataframe containing MPRAGE metadata to obtain last processed image for the specified visit.

    Args:
        mprage_meta_subj: DataFrame of MPRAGE metadata of images corresponding to the subject
        visit_str: Visit name
        unwanted_series_id: Series ID of scans not passing QC

    Returns:
        Dataframe
    """
    mprage_meta_subj = mprage_meta_subj[
        ~mprage_meta_subj.SeriesID.isin(unwanted_series_id)
    ]

    # Get the preferred scan (image series that has been Scaled)
    filtered_mprage = mprage_meta_subj[
        (mprage_meta_subj["Orig/Proc"] == "Processed")
        & (mprage_meta_subj.Visit == visit_str)
        & mprage_meta_subj.Sequence.str.endswith("Scaled", na=False)
    ]

    # If no preferred image found, get N3 processed image (N3m)
    if filtered_mprage.empty:
        filtered_mprage = mprage_meta_subj[
            (mprage_meta_subj["Orig/Proc"] == "Processed")
            & (mprage_meta_subj.Visit == visit_str)
            & mprage_meta_subj.Sequence.str.endswith("N3m", na=False)
        ]

    return filtered_mprage


def adni3_image(subject_id, timepoint, visit_str, mprage_meta_subj, mayo_mri_qc_subj):
    """Select the preferred scan for a subject in a visit, given the subject belongs to ADNI 3 cohort.

    Args:
        subject_id: string containing subject ID
        timepoint: string of visit code in months
        visit_str: string of visit name
        mprage_meta_subj: DataFrame of MPRAGE metadata of images corresponding to the subject
        mayo_mri_qc_subj: DatFrame of MAYO Clinic MR image quality of images corresponding to the subject

    Returns:
        Dictionary containing selected scan information
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        replace_sequence_chars,
    )
    from clinica.utils.stream import cprint

    filtered_scan = mprage_meta_subj[
        (mprage_meta_subj["Orig/Proc"] == "Original")
        & (mprage_meta_subj.Visit == visit_str)
        & mprage_meta_subj.Sequence.str.contains("accel", case=False, na=False)
        & ~mprage_meta_subj.Sequence.str.lower().str.endswith("_nd", na=False)
    ]

    if filtered_scan.empty:
        cprint(
            "NO MPRAGE Meta for ADNI3: "
            + subject_id
            + " for visit "
            + timepoint
            + " - "
            + visit_str
        )
        return None

    scan = select_scan_from_qc(
        filtered_scan, mayo_mri_qc_subj, preferred_field_strength=3.0
    )
    sequence = replace_sequence_chars(scan.Sequence)

    return {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": sequence,
        "Scan_Date": scan.ScanDate,
        "Study_ID": str(scan.StudyID),
        "Series_ID": str(scan.SeriesID),
        "Image_ID": str(scan.ImageUID),
        "Field_Strength": scan.MagStrength,
        "Original": True,
    }


def original_image(
    subject_id,
    timepoint,
    visit_str,
    mprage_meta_subj,
    mayo_mri_qc_subj,
    preferred_field_strength=3.0,
):
    """Select the preferred scan for a subject in a visit, given the scan is not preprocessed.

    Args:
        subject_id:string containing subject ID
        timepoint: string of visit code in months
        visit_str: string of visit name
        mprage_meta_subj: DataFrame of MPRAGE metadata of images corresponding to the subject
        mayo_mri_qc_subj: DatFrame of MAYO Clinic MR image quality of images corresponding to the subject
        preferred_field_strength: Field strength that is preferred in case there are several image acquisitions


    Returns:
        Dictionary containing selected scan information
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        replace_sequence_chars,
    )
    from clinica.utils.stream import cprint

    mprage_meta_subj_orig = mprage_meta_subj[
        mprage_meta_subj["Orig/Proc"] == "Original"
    ]

    cond_mprage = (
        (mprage_meta_subj_orig.Visit == visit_str)
        & mprage_meta_subj_orig.Sequence.str.contains(
            "mprage|mp-rage|mp rage", case=False, na=False
        )
        & ~mprage_meta_subj_orig.Sequence.str.contains("2", na=False)
    )

    cond_spgr = (
        (mprage_meta_subj_orig.Visit == visit_str)
        & mprage_meta_subj_orig.Sequence.str.contains("spgr", case=False, na=False)
        & ~mprage_meta_subj_orig.Sequence.str.contains("acc", case=False, na=False)
    )

    filtered_scan = mprage_meta_subj_orig[cond_mprage | cond_spgr]

    if filtered_scan.empty:
        cprint(
            "NO MPRAGE Meta: "
            + subject_id
            + " for visit "
            + timepoint
            + " - "
            + visit_str
        )
        return None

    scan = select_scan_from_qc(
        filtered_scan, mayo_mri_qc_subj, preferred_field_strength
    )
    sequence = replace_sequence_chars(scan.Sequence)

    return {
        "Subject_ID": subject_id,
        "VISCODE": timepoint,
        "Visit": visit_str,
        "Sequence": sequence,
        "Scan_Date": scan.ScanDate,
        "Study_ID": str(scan.StudyID),
        "Series_ID": str(scan.SeriesID),
        "Image_ID": str(scan.ImageUID),
        "Field_Strength": scan.MagStrength,
        "Original": True,
    }


def select_scan_from_qc(scans_meta, mayo_mri_qc_subj, preferred_field_strength):
    """Select a scan from a list of scans taking into account available QC.

    Args:
        scans_meta: DataFrame containing the metadata for images to choose among
        mayo_mri_qc_subj: DatFrame of MAYO Clinic MR image quality of images corresponding to the subject
        preferred_field_strength: Field strength that is preferred in case there are several image acquisitions

    Returns:
        DataFrame row containing selected scan
    """
    import numpy as np

    multiple_mag_strength = False

    # Select preferred_field_strength images
    if len(scans_meta.MagStrength.unique()) > 1:
        multiple_mag_strength = True
        # Save for later the scans with the non preferred magnetic field strength
        not_preferred_scan = scans_meta[
            scans_meta.MagStrength != preferred_field_strength
        ]
        # Filtering to keep only the scans with the preferred magnetic field strength
        scans_meta = scans_meta[scans_meta.MagStrength == preferred_field_strength]

    if scans_meta.MagStrength.unique()[0] == 3.0:

        id_list = scans_meta.ImageUID.unique()
        image_ids = ["I" + str(imageuid) for imageuid in id_list]
        int_ids = [int(imageuid) for imageuid in id_list]
        images_qc = mayo_mri_qc_subj[mayo_mri_qc_subj.loni_image.isin(image_ids)]

        if not images_qc.empty:
            selected_image = None
            # We check if there is a single selected series image
            if np.sum(images_qc.series_selected) == 1:
                selected_image = (
                    images_qc[images_qc.series_selected == 1].iloc[0].loni_image[1:]
                )
            # Otherwise we select the one with the best QC if available
            else:
                images_not_rejected = images_qc[images_qc.series_quality < 4]

                if images_not_rejected.empty:

                    # There are no images that passed the qc
                    # so we'll try to see if there are other images without qc,
                    # otherwise return None
                    qc_ids = set(
                        [int(qc_id[1:]) for qc_id in images_qc.loni_image.unique()]
                    )
                    no_qc_ids = list(set(int_ids) - qc_ids)

                    # If none of images passed qc the scan is None, otherwise:
                    if no_qc_ids:
                        no_qc_scans_meta = scans_meta[
                            scans_meta.ImageUID.isin(no_qc_ids)
                        ]
                        return select_scan_no_qc(no_qc_scans_meta)
                else:
                    # We select the image with the best (lower) QC.
                    # If no positive QC available we choose image with -1 (not performed)
                    series_quality = [
                        q if q > 0 else 4
                        for q in list(images_not_rejected.series_quality)
                    ]
                    best_q = np.amin(series_quality)
                    if best_q == 4:
                        best_q = -1
                    images_best_qc = images_not_rejected[
                        images_not_rejected.series_quality == best_q
                    ]
                    if len(images_best_qc) == 1:
                        selected_image = images_best_qc.iloc[0].loni_image[1:]
                    else:
                        best_ids = [
                            int(x[1:]) for x in images_best_qc.loni_image.unique()
                        ]
                        best_qc_meta = scans_meta[scans_meta.ImageUID.isin(best_ids)]
                        return select_scan_no_qc(best_qc_meta)

            # If we did not find an image passing QC for the preferred magnetic field strength,
            # then we will choose between the other available images with other magnetic field strength
            if not selected_image and multiple_mag_strength:
                scans_meta = not_preferred_scan
            else:
                scan = scans_meta[scans_meta.ImageUID == int(selected_image)].iloc[0]
                return scan

    # 1.5T (no available QC)
    return select_scan_no_qc(scans_meta)


def select_scan_no_qc(scans_meta):
    """Select a scan from available scans in case there is not an available QC.

    Args:
        scans_meta: DataFrame containing the metadata for images to choose among

    Returns:
        DataFrame row containing selected scan
    """
    # We choose the first scan (not containing repeat in name)
    selected_scan = scans_meta[
        ~scans_meta.Sequence.str.contains("repeat", case=False, na=False)
    ]

    if selected_scan.empty:
        selected_scan = scans_meta

    scan = selected_scan.iloc[0]
    return scan


def check_qc(scan, subject_id, visit_str, mri_quality_subj):
    """Check if a scan has passed check quality, if it was performed.

    Args:
        scan: DataFrame row containing scan information
        subject_id: string containing subject ID
        visit_str: string of visit name
        mri_quality_subj: DatFrame of MR image quality of images corresponding to the subject

    Returns:
        boolean, True if image passed QC or if there is not available QC, False elsewhere
    """
    from clinica.utils.stream import cprint

    # Check if QC exists for image series
    qc = mri_quality_subj[mri_quality_subj.LONIUID == "S" + str(scan.SeriesID)]

    # If QC exists and failed we keep the other scan (in case 2 scans were performed)
    if not qc.empty and qc.iloc[0].PASS != 1:

        cprint("QC found but NOT passed")
        cprint(
            f"Subject {subject_id} for visit {visit_str} "
            f"- Series: {str(scan.SeriesID)} - Study: {str(scan.StudyID)}"
        )
        return False
    return True



# coding: utf-8

"""Module for converting FDG PET of ADNI."""


def convert_adni_fdg_pet(
    source_dir, csv_dir, dest_dir, conversion_dir, subjs_list=None, mod_to_update=False
):
    """Convert FDG PET images of ADNI into BIDS format.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        conversion_dir: path to the TSV files including the paths to original images
        subjs_list: subjects list
        mod_to_update: If True, pre-existing images in the BIDS directory will be erased and extracted again.
    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import paths_to_bids
    from clinica.utils.stream import cprint

    if not subjs_list:
        adni_merge_path = path.join(csv_dir, "ADNIMERGE.csv")
        adni_merge = pd.read_csv(adni_merge_path, sep=",", low_memory=False)
        subjs_list = list(adni_merge.PTID.unique())

    cprint(
        f"Calculating paths of FDG PET images. Output will be stored in {conversion_dir}."
    )
    images = compute_fdg_pet_paths(
        source_dir, csv_dir, dest_dir, subjs_list, conversion_dir
    )
    cprint("Paths of FDG PET images found. Exporting images into BIDS ...")
    paths_to_bids(images, dest_dir, "fdg", mod_to_update=mod_to_update)
    cprint(msg="FDG PET conversion done.", lvl="debug")


def compute_fdg_pet_paths(source_dir, csv_dir, dest_dir, subjs_list, conversion_dir):
    """Compute the paths to the FDG PET images and store them in a TSV file.

    Args:
        source_dir: path to the ADNI directory
        csv_dir: path to the clinical data directory
        dest_dir: path to the destination BIDS directory
        subjs_list: subjects list
        conversion_dir: path to the TSV files including the paths to original images

    Returns:
        images: a dataframe with all the paths to the PET images that will be converted into BIDS
    """
    from os import path

    import pandas as pd

    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        find_image_path,
        get_images_pet,
    )

    pet_fdg_col = [
        "Phase",
        "Subject_ID",
        "VISCODE",
        "Visit",
        "Sequence",
        "Scan_Date",
        "Study_ID",
        "Series_ID",
        "Image_ID",
        "Original",
    ]
    pet_fdg_df = pd.DataFrame(columns=pet_fdg_col)
    pet_fdg_dfs_list = []

    # Loading needed .csv files
    petqc = pd.read_csv(path.join(csv_dir, "PETQC.csv"), sep=",", low_memory=False)
    petqc3 = pd.read_csv(path.join(csv_dir, "PETC3.csv"), sep=",", low_memory=False)
    pet_meta_list = pd.read_csv(
        path.join(csv_dir, "PET_META_LIST.csv"), sep=",", low_memory=False
    )

    for subj in subjs_list:

        # PET images metadata for subject
        subject_pet_meta = pet_meta_list[pet_meta_list["Subject"] == subj]

        if subject_pet_meta.empty:
            continue

        # QC for FDG PET images for ADNI 1, GO and 2
        pet_qc_1go2_subj = petqc[(petqc.PASS == 1) & (petqc.RID == int(subj[-4:]))]

        # QC for FDG PET images for ADNI 3
        pet_qc3_subj = petqc3[(petqc3.SCANQLTY == 1) & (petqc3.RID == int(subj[-4:]))]
        pet_qc3_subj.insert(0, "EXAMDATE", pet_qc3_subj.SCANDATE.to_list())

        # Concatenating visits in both QC files
        pet_qc_subj = pd.concat(
            [pet_qc_1go2_subj, pet_qc3_subj], axis=0, ignore_index=True, sort=False
        )

        sequences_preprocessing_step = ["Co-registered, Averaged"]
        subj_dfs_list = get_images_pet(
            subj,
            pet_qc_subj,
            subject_pet_meta,
            pet_fdg_col,
            "FDG-PET",
            sequences_preprocessing_step,
        )
        if subj_dfs_list:
            pet_fdg_dfs_list += subj_dfs_list

    if pet_fdg_dfs_list:
        # Concatenating dataframes into one
        pet_fdg_df = pd.concat(pet_fdg_dfs_list, ignore_index=True)

    # Exceptions
    # ==========
    conversion_errors = [  # NONAME.nii
        ("031_S_0294", "bl"),
        ("037_S_1421", "m36"),
        ("037_S_1078", "m36"),
        # Empty folders
        ("941_S_1195", "m48"),
        ("005_S_0223", "m12"),
    ]

    # Removing known exceptions from images to convert
    if not pet_fdg_df.empty:
        error_ind = pet_fdg_df.index[
            pet_fdg_df.apply(
                lambda x: ((x.Subject_ID, x.VISCODE) in conversion_errors), axis=1
            )
        ]
        pet_fdg_df.drop(error_ind, inplace=True)

    images = find_image_path(pet_fdg_df, source_dir, "FDG", "I", "Image_ID")
    images.to_csv(path.join(conversion_dir, "fdg_pet_paths.tsv"), sep="\t", index=False)

    return images



# coding: utf-8


def bids_dir_to_fsl_dir(bids_dir):
    """
    Converts BIDS PhaseEncodingDirection parameters (i,j,k,i-,j-,k-) to
    FSL direction (x,y,z,x-,y-,z-).
    """
    fsl_dir = bids_dir.lower()
    if fsl_dir == "i-":
        return "x-"
    if fsl_dir == "i":
        return "x"
    if fsl_dir == "j-":
        return "y-"
    if fsl_dir == "j":
        return "y"
    if fsl_dir == "k-":
        return "z-"
    if fsl_dir == "k":
        return "z"

    raise RuntimeError(
        f"PhaseEncodingDirection {fsl_dir}  is unknown, it should be a value in (x,y,z,x-,y-,z-)"
    )

    return fsl_dir



# coding: utf8


import clinica.pipelines.engine as cpe


class T1FreeSurferLongitudinalCorrection(cpe.Pipeline):
    """FreeSurfer Longitudinal correction class.

    Returns:
        A clinica pipeline object containing the T1FreeSurferLongitudinalCorrection pipeline
    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""

    def check_custom_dependencies(self):
        """Check dependencies that cannot be listed in `info.json`."""

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Note:
            The list of inputs of the T1FreeSurferLongitudinalCorrection pipeline is:
                * participant_id (str): Participant ID (e.g. "sub-CLNC01")
                * session_id (str): Session ID associated to `participant_id` (e.g. "ses-M00")
                * long_id (str): Longitudinal ID associated to `participant_id` (e.g. "long-M00" or "long-M00M18")

        Returns:
            A list of (string) input fields name.
        """
        return ["participant_id", "session_id", "long_id"]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Note:
            The list of inputs of the T1FreeSurferLongitudinalCorrection pipeline is:
                * subject_id (str): FreeSurfer ID
                    (e.g. "sub-CLNC01_ses-M00.long.sub-CLNC01_long-M00M18")

        Returns:
            A list of (string) output fields name.
        """
        return ["subject_id"]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.input_files import T1_FS_DESTRIEUX, T1_FS_T_DESTRIEUX
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint

        from .longitudinal_utils import (
            extract_subject_session_longitudinal_ids_from_filename,
            grab_image_ids_from_caps_directory,
            read_part_sess_long_ids_from_tsv,
            save_part_sess_long_ids_to_tsv,
        )
        from .t1_freesurfer_longitudinal_correction_utils import get_processed_images

        #  We re-initialize subjects, sessions and add long IDS since the latter is not handled by Clinica
        if self.tsv_file:
            (
                self.subjects,
                self.sessions,
                list_long_id,
            ) = read_part_sess_long_ids_from_tsv(self.tsv_file)
        else:
            (
                self.subjects,
                self.sessions,
                list_long_id,
            ) = grab_image_ids_from_caps_directory(self.caps_directory)
        # TODO: Find a way to distinguish cross-sectional from longitudinal pipelines in Clinica Core

        # Display image(s) already present in CAPS folder
        # ===============================================
        processed_ids = get_processed_images(
            self.caps_directory, self.subjects, self.sessions, list_long_id
        )
        if len(processed_ids) > 0:
            cprint(
                msg=f"Clinica found {len(processed_ids)} images(s) already processed in CAPS directory:",
                lvl="warning",
            )
            for image_id in processed_ids:
                cprint(f"{image_id.replace('_', ' | ')}", lvl="warning")
            if self.overwrite_caps:
                output_folder = "<CAPS>/subjects/<participant_id>/<session_id>/t1/<long_id>/freesurfer_longitudinal/"
                cprint(msg=f"Output folders in {output_folder} will be recreated.", lvl="warning")
            else:
                cprint("Image(s) will be ignored by Clinica.", lvl="warning")
                input_ids = [
                    p_id + "_" + s_id + "_" + l_id
                    for p_id, s_id, l_id in zip(
                        self.subjects, self.sessions, list_long_id
                    )
                ]
                to_process_ids = list(set(input_ids) - set(processed_ids))
                (
                    self.subjects,
                    self.sessions,
                    list_long_id,
                ) = extract_subject_session_longitudinal_ids_from_filename(
                    to_process_ids
                )

        all_errors = []
        try:
            # Check that t1-freesurfer has run on the CAPS directory
            clinica_file_reader(
                self.subjects, self.sessions, self.caps_directory, T1_FS_DESTRIEUX
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            # Check that t1-freesurfer-template has run on the CAPS directory
            clinica_file_reader(
                self.subjects, list_long_id, self.caps_directory, T1_FS_T_DESTRIEUX
            )
        except ClinicaException as e:
            all_errors.append(e)

        if len(all_errors) > 0:
            error_message = "Clinica faced errors while trying to read files in your CAPS directory.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaException(error_message)

        # Save subjects to process in <WD>/<Pipeline.name>/participants.tsv
        folder_participants_tsv = os.path.join(self.base_dir, self.name)
        save_part_sess_long_ids_to_tsv(
            self.subjects, self.sessions, list_long_id, folder_participants_tsv
        )

        def print_images_to_process(
            list_participant_id, list_session_id, list_longitudinal_id
        ):
            cprint(
                f"The pipeline will be run on the following {len(list_participant_id)} image(s):"
            )
            for (p_id, s_id, l_id) in zip(
                list_participant_id, list_session_id, list_longitudinal_id
            ):
                cprint(f"\t{p_id} | {s_id} | {l_id}")

        if len(self.subjects):
            # TODO: Generalize long IDs to the message display
            print_images_to_process(self.subjects, self.sessions, list_long_id)
            cprint(
                "List available in %s"
                % os.path.join(folder_participants_tsv, "participants.tsv")
            )
            cprint("The pipeline will last approximately 10 hours per participant.")

        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                ("participant_id", self.subjects),
                ("session_id", self.sessions),
                ("long_id", list_long_id),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        self.connect(
            [
                (read_node, self.input_node, [("participant_id", "participant_id")]),
                (read_node, self.input_node, [("session_id", "session_id")]),
                (read_node, self.input_node, [("long_id", "long_id")]),
            ]
        )

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from .t1_freesurfer_longitudinal_correction_utils import save_to_caps

        save_to_caps = npe.Node(
            interface=nutil.Function(
                input_names=["source_dir", "subject_id", "caps_dir", "overwrite_caps"],
                output_names=["image_id"],
                function=save_to_caps,
            ),
            name="SaveToCaps",
        )
        save_to_caps.inputs.source_dir = os.path.join(
            self.base_dir, self.name, "ReconAll"
        )
        save_to_caps.inputs.caps_dir = self.caps_directory
        save_to_caps.inputs.overwrite_caps = self.overwrite_caps

        self.connect(
            [
                (self.output_node, save_to_caps, [("subject_id", "subject_id")]),
            ]
        )

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import os

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from .t1_freesurfer_longitudinal_correction_utils import (
            init_input_node,
            move_subjects_dir_to_source_dir,
            run_recon_all_long,
            write_tsv_files,
        )

        # Nodes declaration
        # =================
        # Initialize the pipeline
        init_input = npe.Node(
            interface=nutil.Function(
                input_names=[
                    "caps_dir",
                    "participant_id",
                    "session_id",
                    "long_id",
                    "output_dir",
                ],
                output_names=["subjects_dir"],
                function=init_input_node,
            ),
            name="0-InitPipeline",
        )
        init_input.inputs.caps_dir = self.caps_directory
        init_input.inputs.output_dir = os.path.join(
            self.base_dir, self.name, "ReconAll"
        )

        # Run recon-all command
        recon_all = npe.Node(
            interface=nutil.Function(
                input_names=[
                    "subjects_dir",
                    "participant_id",
                    "session_id",
                    "long_id",
                    "directive",
                ],
                output_names=["subject_id"],
                function=run_recon_all_long,
            ),
            name="1-SegmentationReconAll",
        )
        recon_all.inputs.directive = "-all"

        # Generate TSV files containing a summary of the regional statistics
        create_tsv = npe.Node(
            interface=nutil.Function(
                input_names=["subjects_dir", "subject_id"],
                output_names=["subject_id"],
                function=write_tsv_files,
            ),
            name="2-CreateTsvFiles",
        )

        # Move $SUBJECT_DIR to source_dir (1 time point case)
        move_subjects_dir = npe.Node(
            interface=nutil.Function(
                input_names=["subjects_dir", "source_dir", "subject_id"],
                output_names=["subject_id"],
                function=move_subjects_dir_to_source_dir,
            ),
            name="3-MoveSubjectsDir",
        )
        move_subjects_dir.inputs.source_dir = os.path.join(
            self.base_dir, self.name, "ReconAll"
        )

        # Connections
        # ===========
        # fmt: off
        self.connect(
            [
                # Initialize the pipeline
                (self.input_node, init_input, [("participant_id", "participant_id"),
                                               ("session_id", "session_id"),
                                               ("long_id", "long_id")]),
                # Run recon-all command
                (init_input, recon_all, [("subjects_dir", "subjects_dir")]),
                (self.input_node, recon_all, [("participant_id", "participant_id"),
                                              ("session_id", "session_id"),
                                              ("long_id", "long_id")]),
                # Generate TSV files
                (init_input, create_tsv, [("subjects_dir", "subjects_dir")]),
                (recon_all, create_tsv, [("subject_id", "subject_id")]),
                # Move $SUBJECT_DIR to source_dir (macOS case)
                (init_input, move_subjects_dir, [("subjects_dir", "subjects_dir")]),
                (create_tsv, move_subjects_dir, [("subject_id", "subject_id")]),
                # Output node
                (move_subjects_dir, self.output_node, [("subject_id", "subject_id")]),
            ]
        )



# coding: utf8

from nipype import config

import clinica.pipelines.engine as cpe

# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
cfg = dict(execution={'parameterize_dirs': False})
config.update_config(cfg)


class DwiPreprocessingUsingT1(cpe.Pipeline):
    """DWI Preprocessing using T1 image for susceptibility distortion step.

    Warnings:
        - Do not use this pipeline if you have fieldmap data in your dataset.

    Returns:
        A clinica pipeline object containing the DWIPreprocessingUsingT1 pipeline.

    Raises:

    """
    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        from clinica.utils.stream import cprint

        low_bval = self.parameters['low_bval']

        if low_bval < 0:
            raise ValueError(f"The low_bval is negative ({low_bval}): it should be zero or close to zero.")

        if low_bval > 100:
            cprint(f"The low_bval parameter is {low_bval}: it should be zero or close to zero", lvl="warning")

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file.
        """
        pass

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipelines.

        Returns:
            A list of (string) input fields name.
        """
        input_list = ['T1w', 'dwi', 'bvec', 'bval']
        return input_list

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipelines.

        Returns:
            A list of (string) output fields name.
        """
        output_list = ['preproc_dwi', 'preproc_bvec', 'preproc_bval',
                       'b0_mask']

        return output_list

    def build_input_node(self):
        """Build and connect an input node to the pipelines.
        """
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.utils.input_files as input_files
        from clinica.utils.dwi import check_dwi_volume
        from clinica.utils.exceptions import ClinicaBIDSError, ClinicaException
        from clinica.utils.inputs import clinica_file_reader

        all_errors = []
        try:
            t1w_files = clinica_file_reader(self.subjects,
                                            self.sessions,
                                            self.bids_directory,
                                            input_files.T1W_NII)
        except ClinicaException as e:
            all_errors.append(e)
        try:
            dwi_files = clinica_file_reader(self.subjects,
                                            self.sessions,
                                            self.bids_directory,
                                            input_files.DWI_NII)
        except ClinicaException as e:
            all_errors.append(e)

        # bval files
        try:
            bval_files = clinica_file_reader(self.subjects,
                                             self.sessions,
                                             self.bids_directory,
                                             input_files.DWI_BVAL)
        except ClinicaException as e:
            all_errors.append(e)

        # bvec files
        try:
            bvec_files = clinica_file_reader(self.subjects,
                                             self.sessions,
                                             self.bids_directory,
                                             input_files.DWI_BVEC)
        except ClinicaException as e:
            all_errors.append(e)

        if len(all_errors) > 0:
            error_message = 'Clinica faced error(s) while trying to read files in your BIDS directory.\n'
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaBIDSError(error_message)

        # Perform the check after potential issue while reading inputs
        for (dwi, bvec, bval) in zip(dwi_files, bvec_files, bval_files):
            check_dwi_volume(in_dwi=dwi, in_bvec=bvec, in_bval=bval)

        read_input_node = npe.Node(name="LoadingCLIArguments",
                                   interface=nutil.IdentityInterface(
                                       fields=self.get_input_fields(),
                                       mandatory_inputs=True),
                                   iterables=[('T1w', t1w_files),
                                              ('dwi', dwi_files),
                                              ('bvec', bvec_files),
                                              ('bval', bval_files)],
                                   synchronize=True)

        self.connect([
            (read_input_node, self.input_node, [('T1w', 'T1w')]),
            (read_input_node, self.input_node, [('dwi', 'dwi')]),
            (read_input_node, self.input_node, [('bvec', 'bvec')]),
            (read_input_node, self.input_node, [('bval', 'bval')])
        ])

    def build_output_node(self):
        """Build and connect an output node to the pipelines.
        """
        import nipype.interfaces.io as nio
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_utils as utils
        from clinica.utils.nipype import fix_join

        # Find container path from DWI filename
        # =====================================
        container_path = npe.Node(nutil.Function(
            input_names=['bids_dwi_filename'],
            output_names=['container'],
            function=utils.dwi_container_from_filename),
            name='container_path')

        rename_into_caps = npe.Node(nutil.Function(
            input_names=['in_bids_dwi', 'fname_dwi', 'fname_bval',
                         'fname_bvec', 'fname_brainmask'],
            output_names=['out_caps_dwi', 'out_caps_bval', 'out_caps_bvec',
                          'out_caps_brainmask'],
            function=utils.rename_into_caps),
            name='rename_into_caps')

        # Writing results into CAPS
        # =========================
        write_results = npe.Node(name='write_results',
                                 interface=nio.DataSink())
        write_results.inputs.base_directory = self.caps_directory
        write_results.inputs.parameterization = False

        self.connect([
            (self.input_node, container_path,    [('dwi',                    'bids_dwi_filename')]),  # noqa
            (self.input_node,  rename_into_caps, [('dwi',                          'in_bids_dwi')]),  # noqa
            (self.output_node, rename_into_caps, [('preproc_dwi',                      'fname_dwi'),  # noqa
                                                  ('preproc_bval',                    'fname_bval'),  # noqa
                                                  ('preproc_bvec',                    'fname_bvec'),  # noqa
                                                  ('b0_mask',                  'fname_brainmask')]),  # noqa
            (container_path, write_results,      [(('container', fix_join, 'dwi'), 'container')]),  # noqa
            (rename_into_caps, write_results,    [('out_caps_dwi',    'preprocessing.@preproc_dwi'),  # noqa
                                                  ('out_caps_bval',  'preprocessing.@preproc_bval'),  # noqa
                                                  ('out_caps_bvec',  'preprocessing.@preproc_bvec'),  # noqa
                                                  ('out_caps_brainmask', 'preprocessing.@b0_mask')])  # noqa
        ])

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipelines.
        """
        import nipype.interfaces.fsl as fsl
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_utils as utils
        import clinica.pipelines.dwi_preprocessing_using_t1.dwi_preprocessing_using_t1_workflows as workflows
        from clinica.utils.dwi import prepare_reference_b0
        from clinica.workflows.dwi_preprocessing import (
            ecc_pipeline,
            hmc_pipeline,
            remove_bias,
        )

        # Nodes creation
        # ==============
        # Prepare b0 image for further corrections
        prepare_b0 = npe.Node(name="PrepareB0", interface=nutil.Function(
            input_names=['in_dwi', 'in_bval', 'in_bvec', 'low_bval'],
            output_names=['out_reference_b0', 'out_b0_dwi_merge',
                          'out_updated_bval', 'out_updated_bvec'],
            function=prepare_reference_b0))
        prepare_b0.inputs.low_bval = self.parameters['low_bval']
        # Mask b0 for computations purposes
        mask_b0_pre = npe.Node(fsl.BET(frac=0.3, mask=True, robust=True),
                               name='PreMaskB0')
        # Head-motion correction
        hmc = hmc_pipeline(name='HeadMotionCorrection')
        # Eddy-currents correction
        ecc = ecc_pipeline(name='EddyCurrentCorrection')
        # Susceptibility distortion correction using T1w image
        sdc = workflows.susceptibility_distortion_correction_using_t1(
            name='SusceptibilityDistortionCorrection')
        # Remove bias correction
        bias = remove_bias(name='RemoveBias')
        # Apply all corrections
        aac = workflows.apply_all_corrections_using_ants(
            name='ApplyAllCorrections')

        print_begin_message = npe.Node(
            interface=nutil.Function(
                input_names=['in_bids_or_caps_file'],
                function=utils.print_begin_pipeline),
            name='Write-Begin_Message')

        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=['in_bids_or_caps_file', 'final_file'],
                function=utils.print_end_pipeline),
            name='Write-End_Message')

        # Connection
        # ==========

        self.connect([
            # Print begin message
            (self.input_node, print_begin_message, [('dwi', 'in_bids_or_caps_file')]),  # noqa
            # Preliminary step (possible computation of a mean b0):
            (self.input_node, prepare_b0, [('dwi',  'in_dwi'),  # noqa
                                           ('bval', 'in_bval'),  # noqa
                                           ('bvec', 'in_bvec')]),  # noqa
            # Mask b0 before corrections
            (prepare_b0, mask_b0_pre, [('out_reference_b0', 'in_file')]),  # noqa
            # Head-motion correction
            (prepare_b0,  hmc, [('out_b0_dwi_merge', 'inputnode.in_file'),  # noqa
                               ('out_updated_bval',  'inputnode.in_bval'),  # noqa
                               ('out_updated_bvec',  'inputnode.in_bvec')]),  # noqa
            (mask_b0_pre, hmc, [('mask_file',        'inputnode.in_mask')]),  # noqa
            # Eddy-current correction
            (hmc,         ecc, [('outputnode.out_xfms', 'inputnode.in_xfms')]),  # noqa
            (prepare_b0,  ecc, [('out_b0_dwi_merge',    'inputnode.in_file')]),  # noqa
            (prepare_b0,  ecc, [('out_updated_bval',    'inputnode.in_bval')]),  # noqa
            (mask_b0_pre, ecc, [('mask_file',           'inputnode.in_mask')]),  # noqa
            # Magnetic susceptibility correction
            (self.input_node, sdc, [('T1w',                 'inputnode.in_t1')]),  # noqa
            (ecc,             sdc, [('outputnode.out_file', 'inputnode.in_dwi')]),  # noqa
            (hmc,             sdc, [('outputnode.out_bvec', 'inputnode.in_bvec')]),  # noqa
            # Apply all corrections
            (prepare_b0,      aac, [('out_b0_dwi_merge',    'inputnode.in_dwi')]),  # noqa
            (hmc,             aac, [('outputnode.out_xfms', 'inputnode.in_hmc')]),  # noqa
            (ecc,             aac, [('outputnode.out_xfms', 'inputnode.in_ecc')]),  # noqa
            (sdc,             aac, [('outputnode.out_warp', 'inputnode.in_sdc_syb')]),  # noqa
            (self.input_node, aac, [('T1w',                 'inputnode.in_t1')]),  # noqa
            # Bias correction
            (aac, bias, [('outputnode.out_file', 'inputnode.in_file')]),
            # Outputnode:
            (bias,       self.output_node, [('outputnode.out_file', 'preproc_dwi')]),  # noqa
            (sdc,        self.output_node, [('outputnode.out_bvec', 'preproc_bvec')]),  # noqa
            (prepare_b0, self.output_node, [('out_updated_bval',    'preproc_bval')]),  # noqa
            (bias,       self.output_node, [('outputnode.b0_mask',  'b0_mask')]),  # noqa
            # Print end message
            (self.input_node, print_end_message, [('dwi',         'in_bids_or_caps_file')]),  # noqa
            (bias,            print_end_message, [('outputnode.out_file', 'final_file')]),  # noqa
        ])



# coding: utf8

"""Methods to find information in the different pipelines of Clinica."""


import os
from glob import glob
from os import path

import pandas as pd


def pet_volume_pipeline(
    caps_dir,
    df,
    group_selection=None,
    volume_atlas_selection=None,
    pvc_restriction=None,
    tracers_selection=None,
    **kwargs,
):
    """Merge the data of the PET-Volume pipeline to the merged file containing the BIDS information.

    Args:
        caps_dir: the path to the CAPS directory
        df: the DataFrame containing the BIDS information
        group_selection: allows to choose the DARTEL groups to merge. If None all groups are selected.
        volume_atlas_selection: allows to choose the atlas to merge (default = 'all')
        pvc_restriction: gives the restriction on the inclusion or not of the file with the label 'pvc-rbv'
            1       --> only the atlases containing the label will be used
            0       --> the atlases containing the label won't be used
            None    --> all the atlases will be used
        tracers_selection: allows to choose the PET tracer to merge (default = 'all')

    Returns:
         final_df: a DataFrame containing the information of the bids and the pipeline
    """
    pet_path = path.join("pet", "preprocessing")

    return volume_pipeline(
        caps_dir,
        df,
        pet_path,
        group_selection=group_selection,
        atlas_selection=volume_atlas_selection,
        pvc_restriction=pvc_restriction,
        pipeline_name="pet-volume",
        tracers_selection=tracers_selection,
    )


def t1_freesurfer_pipeline(caps_dir, df, freesurfer_atlas_selection=None, **kwargs):
    """Merge the data of the PET-Volume pipeline to the merged file containing the BIDS information.

    Args:
        caps_dir: the path to the CAPS directory
        df: the DataFrame containing the BIDS information
        freesurfer_atlas_selection: allows to choose the atlas to merge (default = 'all')

    Returns:
         final_df: a DataFrame containing the information of the bids and the pipeline
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        replace_sequence_chars,
    )
    from clinica.utils.stream import cprint

    # Ensures that df is correctly indexed
    if "participant_id" in df.columns.values:
        df.set_index(["participant_id", "session_id"], inplace=True, drop=True)

    subjects_dir = path.join(caps_dir, "subjects")

    pipeline_df = pd.DataFrame()

    for participant_id, session_id in df.index.values:

        ses_path = path.join(subjects_dir, participant_id, session_id)
        mod_path = path.join(
            ses_path, "t1", "freesurfer_cross_sectional", "regional_measures"
        )
        ses_df = pd.DataFrame(
            [[participant_id, session_id]], columns=["participant_id", "session_id"]
        )
        ses_df.set_index(["participant_id", "session_id"], inplace=True, drop=True)

        if os.path.exists(mod_path):
            # Looking for atlases
            atlas_paths = glob(
                path.join(mod_path, f"{participant_id}_{session_id}_*thickness.tsv")
            )

            for atlas_path in atlas_paths:
                atlas_name = atlas_path.split("_parcellation-")[1].split("_")[0]
                if path.exists(atlas_path) and (
                    not freesurfer_atlas_selection
                    or (
                        freesurfer_atlas_selection
                        and atlas_name in freesurfer_atlas_selection
                    )
                ):
                    atlas_df = pd.read_csv(atlas_path, sep="\t")
                    label_list = [
                        f"t1-freesurfer_atlas-{atlas_name}_ROI-{replace_sequence_chars(roi_name)}_thickness"
                        for roi_name in atlas_df.label_name.values
                    ]
                    ses_df[label_list] = atlas_df["label_value"].to_numpy()

            # Always retrieve subcortical volumes
            atlas_path = path.join(
                mod_path, f"{participant_id}_{session_id}_segmentationVolumes.tsv"
            )
            atlas_df = pd.read_csv(atlas_path, sep="\t")
            label_list = [
                f"t1-freesurfer_segmentation-volumes_ROI-{replace_sequence_chars(roi_name)}_volume"
                for roi_name in atlas_df.label_name.values
            ]
            ses_df[label_list] = atlas_df["label_value"].to_numpy()

        pipeline_df = pipeline_df.append(ses_df)

    summary_df = generate_summary(pipeline_df, "t1-freesurfer", ignore_groups=True)
    final_df = pd.concat([df, pipeline_df], axis=1)
    final_df.reset_index(inplace=True)

    return final_df, summary_df


def t1_volume_pipeline(
    caps_dir, df, group_selection=None, volume_atlas_selection=None, **kwargs
):
    """Merge data of the t1-volume pipeline to the merged file containing the BIDS information.

    Args:
        caps_dir: the path to the CAPS directory
        df: the DataFrame containing the BIDS information
        group_selection: allows to choose the DARTEL groups to merge. If None all groups are selected.
        volume_atlas_selection: allows to choose the atlas to merge. If None all atlases are selected.

    Returns:
        final_df: a DataFrame containing the information of the bids and the pipeline
    """
    t1_spm_path = path.join("t1", "spm", "dartel")

    return volume_pipeline(
        caps_dir,
        df,
        t1_spm_path,
        group_selection=group_selection,
        atlas_selection=volume_atlas_selection,
        pvc_restriction=None,
        pipeline_name="t1-volume",
    )


def volume_pipeline(
    caps_dir,
    df,
    pipeline_path,
    pipeline_name,
    group_selection=None,
    atlas_selection=None,
    pvc_restriction=None,
    tracers_selection=None,
):
    """Merge data of the t1-volume and pet-volume pipelines to the merged file containing the BIDS information.

    Args:
        caps_dir: the path to the CAPS directory
        df: the DataFrame containing the BIDS information
        pipeline_path: path between the session folder and the group folder
        pipeline_name: name of the pipeline
        group_selection: allows to choose the DARTEL groups to merge. If None all groups are selected.
        atlas_selection: allows to choose the atlas to merge. If None all atlases are selected.
        pvc_restriction: gives the restriction on the inclusion or not of the file with the label 'pvc-rbv'
            1       --> only the atlases containing the label will be used
            0       --> the atlases containing the label won't be used
            None    --> all the atlases will be used
        tracers_selection: allows to choose the PET tracer to merge (default = 'all')

    Returns:
        final_df: a DataFrame containing the information of the bids and the pipeline
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        replace_sequence_chars,
    )

    # Ensures that df is correctly indexed
    if "participant_id" in df.columns.values:
        df.set_index(["participant_id", "session_id"], inplace=True, drop=True)

    if not group_selection:
        try:
            group_selection = os.listdir(path.join(caps_dir, "groups"))
        except FileNotFoundError:
            return df, None
    else:
        group_selection = [f"group-{group}" for group in group_selection]

    subjects_dir = path.join(caps_dir, "subjects")

    pipeline_df = pd.DataFrame()

    for participant_id, session_id in df.index.values:

        ses_path = path.join(subjects_dir, participant_id, session_id)
        mod_path = path.join(ses_path, pipeline_path)
        ses_df = pd.DataFrame(
            [[participant_id, session_id]], columns=["participant_id", "session_id"]
        )
        ses_df.set_index(["participant_id", "session_id"], inplace=True, drop=True)

        if os.path.exists(mod_path):
            # Looking for groups
            for group in group_selection:
                group_path = path.join(mod_path, group)
                if os.path.exists(group_path):
                    # Looking for atlases
                    if not atlas_selection:
                        atlas_paths = glob(
                            path.join(
                                group_path,
                                "atlas_statistics",
                                f"{participant_id}_{session_id}_*_statistics.tsv",
                            )
                        )
                    else:
                        atlas_paths = list()
                        for atlas in atlas_selection:
                            atlas_paths += glob(
                                path.join(
                                    group_path,
                                    "atlas_statistics",
                                    f"{participant_id}_{session_id}_*{atlas}*_statistics.tsv",
                                )
                            )

                    # Filter pvc_restriction
                    if pvc_restriction:
                        if pvc_restriction == 1:
                            atlas_paths = [
                                atlas_path
                                for atlas_path in atlas_paths
                                if "pvc-rbv" in atlas_path
                            ]
                        else:
                            atlas_paths = [
                                atlas_path
                                for atlas_path in atlas_paths
                                if "pvc-rbv" not in atlas_path
                            ]

                    # Filter tracers
                    if tracers_selection:
                        atlas_paths = [
                            atlas_path
                            for atlas_path in atlas_paths
                            for tracer in tracers_selection
                            if tracer in atlas_path
                        ]

                    for atlas_path in atlas_paths:
                        atlas_name = atlas_path.split("_space-")[-1].split("_")[0]
                        if path.exists(atlas_path):
                            atlas_df = pd.read_csv(atlas_path, sep="\t")
                            additional_desc = ""
                            if "acq" in atlas_path:
                                tracer = atlas_path.split("_acq-")[1].split("_")[0]
                                additional_desc += f"_acq-{tracer}"
                            if "pvc-rbv" in atlas_path:
                                additional_desc += f"_pvc-rbv"

                            label_list = [
                                f"{pipeline_name}_{group}_atlas-{atlas_name}{additional_desc}_ROI-{replace_sequence_chars(roi_name)}_intensity"
                                for roi_name in atlas_df.label_name.values
                            ]
                            ses_df[label_list] = atlas_df["mean_scalar"].to_numpy()

        pipeline_df = pipeline_df.append(ses_df)

    summary_df = generate_summary(pipeline_df, pipeline_name)
    final_df = pd.concat([df, pipeline_df], axis=1)
    final_df.reset_index(inplace=True)

    return final_df, summary_df


def generate_summary(pipeline_df, pipeline_name, ignore_groups=False):

    columns = [
        "pipeline_name",
        "group_id",
        "atlas_id",
        "tracer",
        "pvc",
        "regions_number",
        "first_column_name",
        "last_column_name",
    ]
    summary_df = pd.DataFrame(columns=columns)

    if ignore_groups:
        groups = ["_"]
        atlases = list({column.split("_")[1] for column in pipeline_df.columns.values})
    else:
        groups = list({column.split("_")[1] for column in pipeline_df.columns.values})
        atlases = list({column.split("_")[2] for column in pipeline_df.columns.values})
    pvc_rectrictions = list(
        {"pvc-rbv" in column for column in pipeline_df.columns.values}
    )
    tracers = list(
        {
            column.split("_acq-")[1].split("_")[0]
            for column in pipeline_df.columns.values
            if "acq" in column
        }
    )
    if len(tracers) == 0:
        tracers.append("_")
        pvc_rectrictions = ["_"]
    groups.sort()
    atlases.sort()

    for group in groups:
        group_id = group.split("-")[-1]
        for atlas in atlases:
            atlas_id = atlas.split("-")[-1]
            for tracer in tracers:
                for pvc_rectriction in pvc_rectrictions:
                    if pvc_rectriction and pvc_rectriction != "_":
                        regions = [
                            column
                            for column in pipeline_df.columns.values
                            if group in column
                            and atlas in column
                            and "pvc-rbv" in column
                            and tracer in column
                        ]
                    else:
                        regions = [
                            column
                            for column in pipeline_df.columns.values
                            if group in column
                            and atlas in column
                            and "pvc-rbv" not in column
                            and tracer in column
                        ]

                    if len(regions) > 0:
                        row_df = pd.DataFrame(
                            [
                                [
                                    pipeline_name,
                                    group_id,
                                    atlas_id,
                                    tracer,
                                    pvc_rectriction,
                                    len(regions),
                                    regions[0],
                                    regions[-1],
                                ]
                            ],
                            columns=columns,
                        )
                        summary_df = summary_df.append(row_df, ignore_index=True)

    summary_df = summary_df.replace("_", "n/a")
    return summary_df


class DatasetError(Exception):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return repr("Bad format for the sessions: " + self.name)



# -*- coding: utf-8 -*-


# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
from nipype import config

import clinica.pipelines.engine as cpe

cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class DeepLearningPrepareData(cpe.Pipeline):
    """Deeplearning prepare data - MRI in nifti format are transformed into
    PyTorch tensors. The transformation is applied to: the whole volume, a
    selection of 3D patches, or slices extracted from the 3D volume. By default
    it uses the cropped version of the MRI (see option "--use_uncropper_image")


    Returns:
        A clinica pipeline object containing the Deeplearning prepare data pipeline.
    """

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """
        return ["input_nifti"]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline.

        Returns:
            A list of (string) output fields name.
        """
        return ["image_id"]

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        from os import path

        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import (
            ClinicaBIDSError,
            ClinicaCAPSError,
            ClinicaException,
        )
        from clinica.utils.input_files import (
            T1W_EXTENSIVE,
            T1W_LINEAR,
            T1W_LINEAR_CROPPED,
            pet_linear_nii,
        )
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        from .deeplearning_prepare_data_utils import check_mask_list

        # Select the correct filetype corresponding to modality
        if self.parameters.get("modality") == "t1-linear":
            if self.parameters.get("use_uncropped_image"):
                FILE_TYPE = T1W_LINEAR
            else:
                FILE_TYPE = T1W_LINEAR_CROPPED
        if self.parameters.get("modality") == "t1-extensive":
            FILE_TYPE = T1W_EXTENSIVE

        if self.parameters.get("modality") == "pet-linear":
            FILE_TYPE = pet_linear_nii(
                self.parameters.get("acq_label"),
                self.parameters.get("suvr_reference_region"),
                self.parameters.get("use_uncropped_image"),
            )
        if self.parameters.get("modality") == "custom":
            FILE_TYPE = {
                "pattern": f"*{self.parameters.get('custom_suffix')}",
                "description": "Custom suffix",
            }

        # Input file:
        try:
            input_files = clinica_file_reader(
                self.subjects, self.sessions, self.caps_directory, FILE_TYPE
            )
        except ClinicaException as e:
            err = (
                "Clinica faced error(s) while trying to read files in your CAPS directory.\n"
                + str(e)
            )
            raise ClinicaBIDSError(err)

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint("The pipeline will last approximately 30 seconds per image.")

        if self.parameters.get("extract_method") == "slice":
            self.slice_direction = self.parameters.get("slice_direction")
            self.slice_mode = self.parameters.get("slice_mode")
        else:
            self.slice_direction = "axial"
            self.slice_mode = "rgb"

        if self.parameters.get("extract_method") == "patch":
            self.patch_size = self.parameters.get("patch_size")
            self.stride_size = self.parameters.get("stride_size")
        else:
            self.patch_size = 50
            self.stride_size = 50

        # Load the corresponding masks
        if self.parameters.get("extract_method") == "roi":
            self.roi_list = self.parameters.get("roi_list")

            if self.parameters.get("modality") == "custom":
                self.mask_pattern = self.parameters.get("custom_mask_pattern")
                self.template = self.parameters.get("custom_template")
                if not self.template:
                    raise ValueError(
                        "A custom template must be defined when the modality is set to custom."
                    )
            else:
                self.mask_pattern = None
                from .deeplearning_prepare_data_utils import TEMPLATE_DICT

                self.template = TEMPLATE_DICT[self.parameters.get("modality")]

            self.masks_location = path.join(
                self.caps_directory, "masks", f"tpl-{self.template}"
            )

            if not self.roi_list:
                raise ValueError("A list of regions must be given.")
            else:
                check_mask_list(
                    self.masks_location,
                    self.roi_list,
                    self.mask_pattern,
                    not self.parameters.get("use_uncropped_image"),
                )
        else:
            self.masks_location = ""
            self.mask_pattern = None
            self.roi_list = []

        # The reading node
        # -------------------------
        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                ("input_nifti", input_files),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )

        self.connect(
            [
                (read_node, self.input_node, [("input_nifti", "input_nifti")]),
            ]
        )

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces.io import DataSink

        from clinica.utils.filemanip import get_subject_id
        from clinica.utils.nipype import container_from_filename, fix_join

        # Write node
        # ----------------------
        write_node = npe.Node(name="WriteCaps", interface=DataSink())
        write_node.inputs.base_directory = self.caps_directory
        write_node.inputs.parameterization = False

        # Get subject ID node
        # ----------------------
        image_id_node = npe.Node(
            interface=nutil.Function(
                input_names=["bids_or_caps_file"],
                output_names=["image_id"],
                function=get_subject_id,
            ),
            name="ImageID",
        )

        # Find container path from input filename
        # ----------------------
        container_path = npe.Node(
            nutil.Function(
                input_names=["bids_or_caps_filename"],
                output_names=["container"],
                function=container_from_filename,
            ),
            name="ContainerPath",
        )

        # fmt: off
        self.connect(
            [
                (self.input_node, image_id_node, [("input_nifti", "bids_or_caps_file")]),
                (self.input_node, container_path, [("input_nifti", "bids_or_caps_filename")]),
                # (image_id_node, write_node, [('image_id', '@image_id')]),
                (image_id_node, write_node, [("image_id", "@image_id")]),
            ]
        )
        # fmt: on

        subfolder = "image_based"
        if self.parameters.get("extract_method") == "slice":
            subfolder = "slice_based"
            # fmt: off
            self.connect(
                [
                    (self.output_node, write_node, [("slices_rgb_output", "@slices_rgb_output")]),
                    (self.output_node, write_node, [("slices_original_output", "@slices_original_output")]),
                ]
            )
            # fmt: on

        elif self.parameters.get("extract_method") == "patch":
            subfolder = "patch_based"
            # fmt: off
            self.connect(
                [
                    (self.output_node, write_node, [("patches_output", "@patches_output")])
                ]
            )
            # fmt: on
        elif self.parameters.get("extract_method") == "roi":
            subfolder = "roi_based"
            # fmt: off
            self.connect(
                [
                    (self.output_node, write_node, [("roi_output", "@roi_output")])
                ]
            )
            # fmt: on
        else:
            # fmt: off
            self.connect(
                [
                    (self.output_node, write_node, [("output_pt_file", "@output_pt_file")])
                ]
            )
            # fmt: on

        mod_subfolder = ""
        if self.parameters.get("modality") == "t1-linear":
            mod_subfolder = "t1_linear"
        if self.parameters.get("modality") == "t1-extensive":
            mod_subfolder = "t1_extensive"
        if self.parameters.get("modality") == "pet-linear":
            mod_subfolder = "pet_linear"
        if self.parameters.get("modality") == "custom":
            mod_subfolder = "custom"

        # fmt: off
        self.connect(
            [
                (container_path, write_node, [
                    (("container", fix_join, "deeplearning_prepare_data", subfolder, mod_subfolder), "container")]),
            ]
        )
        # fmt: on

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from .deeplearning_prepare_data_utils import (
            extract_patches,
            extract_roi,
            extract_slices,
            save_as_pt,
        )

        # The processing nodes
        # Node to save input in nii.gz format into pytorch .pt format
        # ----------------------
        save_as_pt = npe.MapNode(
            name="save_as_pt",
            iterfield=["input_img"],
            interface=nutil.Function(
                function=save_as_pt,
                input_names=["input_img"],
                output_names=["output_file"],
            ),
        )

        # Extract slices node (options: 3 directions, mode)
        # ----------------------
        extract_slices_node = npe.MapNode(
            name="extract_slices",
            iterfield=["input_tensor"],
            interface=nutil.Function(
                function=extract_slices,
                input_names=["input_tensor", "slice_direction", "slice_mode"],
                output_names=["output_file_rgb", "output_file_original"],
            ),
        )

        extract_slices_node.inputs.slice_direction = self.slice_direction
        extract_slices_node.inputs.slice_mode = self.slice_mode

        # Extract patches node (options, patch size and stride size)
        # ----------------------
        extract_patches_node = npe.MapNode(
            name="extract_patches",
            iterfield=["input_tensor"],
            interface=nutil.Function(
                function=extract_patches,
                input_names=["input_tensor", "patch_size", "stride_size"],
                output_names=["output_patch"],
            ),
        )

        extract_patches_node.inputs.patch_size = self.patch_size
        extract_patches_node.inputs.stride_size = self.stride_size

        # Extract ROi node
        extract_roi_node = npe.MapNode(
            name="extract_ROI",
            iterfield=["input_tensor"],
            interface=nutil.Function(
                function=extract_roi,
                input_names=[
                    "input_tensor",
                    "masks_location",
                    "mask_pattern",
                    "cropped_input",
                    "roi_list",
                    "uncrop_output",
                ],
                output_names=["output_roi"],
            ),
        )
        extract_roi_node.inputs.masks_location = self.masks_location
        extract_roi_node.inputs.mask_pattern = self.mask_pattern
        extract_roi_node.inputs.cropped_input = not self.parameters.get(
            "use_uncropped_image"
        )
        extract_roi_node.inputs.roi_list = self.roi_list
        extract_roi_node.inputs.uncrop_output = self.parameters.get("roi_uncrop_output")

        # Connections
        # ----------------------
        # fmt: off
        self.connect(
            [
                (self.input_node, save_as_pt, [("input_nifti", "input_img")]),
            ]
        )

        if self.parameters.get("extract_method") == "slice":
            self.connect(
                [
                    (save_as_pt, extract_slices_node, [("output_file", "input_tensor")]),
                    (extract_slices_node, self.output_node, [("output_file_rgb", "slices_rgb_output")]),
                    (extract_slices_node, self.output_node, [("output_file_original", "slices_original_output")]),
                ]
            )
        elif self.parameters.get("extract_method") == "patch":
            self.connect(
                [
                    (save_as_pt, extract_patches_node, [("output_file", "input_tensor")]),
                    (extract_patches_node, self.output_node, [("output_patch", "patches_output")]),
                ]
            )
        elif self.parameters.get("extract_method") == "roi":
            self.connect(
                [
                    (save_as_pt, extract_roi_node, [("output_file", "input_tensor")]),
                    (extract_roi_node, self.output_node, [("output_roi", "roi_output")]),
                ]
            )
        else:
            self.connect(
                [
                    (save_as_pt, self.output_node, [("output_file", "output_pt_file")]),
                ]
            )
        # fmt: on



# -*- coding: utf-8 -*-
# emacs: -*- mode: python; py-indent-offset: 4; indent-tabs-mode: nil -*-
# vi: set ft=python sts=4 ts=4 sw=4 et:
from __future__ import absolute_import, division, print_function, unicode_literals

import os.path as op

from nipype.interfaces.base import (
    CommandLine,
    CommandLineInputSpec,
    File,
    InputMultiPath,
    TraitedSpec,
    isdefined,
    traits,
)
from nipype.utils.filemanip import split_filename


class DWI2TensorInputSpec(CommandLineInputSpec):
    in_file = InputMultiPath(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Diffusion-weighted images",
    )
    out_filename = File(
        name_template="%s_tensor.mif",
        name_source="in_file",
        output_name="tensor",
        argstr="%s",
        desc="Output tensor filename",
        position=-1,
    )
    encoding_file = File(
        argstr="-grad %s",
        position=2,
        desc=(
            "Encoding file supplied as a 4xN text file with "
            "each line is in the format [ X Y Z b ], where "
            "[ X Y Z ] describe the direction of the applied "
            "gradient, and b gives the b-value in units "
            "(1000 s/mm^2). See FSL2MRTrix()"
        ),
    )
    ignore_slice_by_volume = traits.List(
        traits.Int,
        argstr="-ignoreslices %s",
        sep=" ",
        position=2,
        minlen=2,
        maxlen=2,
        desc=(
            "Requires two values (i.e. [34 "
            "1] for [Slice Volume] Ignores "
            "the image slices specified "
            "when computing the tensor. "
            "Slice here means the z "
            "coordinate of the slice to be "
            "ignored."
        ),
    )
    ignore_volumes = traits.List(
        traits.Int,
        argstr="-ignorevolumes %s",
        sep=" ",
        position=2,
        minlen=1,
        desc=(
            "Requires two values (i.e. [2 5 6] for "
            "[Volumes] Ignores the image volumes "
            "specified when computing the tensor."
        ),
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc=("Do not display information messages or progress " "status."),
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")
    in_mask = File(
        exists=True,
        argstr="-mask %s",
        desc=(
            "only perform computation within the specified binary" " brain mask image"
        ),
    )


class DWI2TensorOutputSpec(TraitedSpec):
    tensor = File(exists=True, desc="path/name of output diffusion tensor image")


class DWI2Tensor(CommandLine):
    """
    Converts diffusion-weighted images to tensor images.

    Example
    -------

    >>> import nipype.interfaces.mrtrix as mrt
    >>> dwi2tensor = mrt.DWI2Tensor()
    >>> dwi2tensor.inputs.in_file = 'dwi.mif'
    >>> dwi2tensor.inputs.encoding_file = 'encoding.txt'
    >>> dwi2tensor.cmdline
    'dwi2tensor -grad encoding.txt dwi.mif dwi_tensor.mif'
    >>> dwi2tensor.run()                                   # doctest: +SKIP
    """

    _cmd = "dwi2tensor"
    input_spec = DWI2TensorInputSpec
    output_spec = DWI2TensorOutputSpec


class MRTransformInputSpec(CommandLineInputSpec):
    in_files = InputMultiPath(
        File(exists=True),
        argstr="%s",
        mandatory=True,
        position=-2,
        desc="Input images to be transformed",
    )
    out_filename = File(genfile=True, argstr="%s", position=-1, desc="Output image")
    invert = traits.Bool(
        argstr="-inverse",
        position=1,
        desc="Invert the specified transform before using it",
    )
    replace_transform = traits.Bool(
        argstr="-replace",
        position=1,
        desc="replace the current transform by that specified, rather than applying it to the current transform",
    )
    transformation_file = File(
        exists=True,
        argstr="-transform %s",
        position=1,
        desc="The transform to apply, in the form of a 4x4 ascii file.",
    )
    linear_transform = File(
        exists=True,
        argstr="-linear %s",
        position=1,
        desc=(
            "specify a linear transform to apply, in the form of a 3x4 or "
            "4x4 ascii file. Note the standard reverse convention is used, "
            "where the transform maps points in the template image to the "
            "moving image. Note that the reverse convention is still assumed "
            "even if no -template image is supplied"
        ),
    )
    template_image = File(
        exists=True,
        argstr="-template %s",
        position=1,
        desc="Reslice the input image to match the specified template image.",
    )
    reference_image = File(
        exists=True,
        argstr="-reference %s",
        position=1,
        desc="in case the transform supplied maps from the input image onto a reference image, use this option to specify the reference. Note that this implicitly sets the -replace option.",
    )
    flip_x = traits.Bool(
        argstr="-flipx",
        position=1,
        desc="assume the transform is supplied assuming a coordinate system with the x-axis reversed relative to the MRtrix convention (i.e. x increases from right to left). This is required to handle transform matrices produced by FSL's FLIRT command. This is only used in conjunction with the -reference option.",
    )
    quiet = traits.Bool(
        argstr="-quiet",
        position=1,
        desc="Do not display information messages or progress status.",
    )
    debug = traits.Bool(argstr="-debug", position=1, desc="Display debugging messages.")


class MRTransformOutputSpec(TraitedSpec):
    out_file = File(exists=True, desc="the output image of the transformation")


class MRTransform(CommandLine):
    """
    Apply spatial transformations or reslice images

    Example
    -------

    >>> MRxform = MRTransform()
    >>> MRxform.inputs.in_files = 'anat_coreg.mif'
    >>> MRxform.run()                                   # doctest: +SKIP
    """

    _cmd = "mrtransform"
    input_spec = MRTransformInputSpec
    output_spec = MRTransformOutputSpec

    def _list_outputs(self):
        outputs = self.output_spec().get()
        outputs["out_file"] = self.inputs.out_filename
        if not isdefined(outputs["out_file"]):
            outputs["out_file"] = op.abspath(self._gen_outfilename())
        else:
            outputs["out_file"] = op.abspath(outputs["out_file"])
        return outputs

    def _gen_filename(self, name):
        if name == "out_filename":
            return self._gen_outfilename()
        else:
            return None

    def _gen_outfilename(self):
        _, name, _ = split_filename(self.inputs.in_files[0])
        return name + "_MRTransform.mif"



# coding: utf8


import numpy as np
from sklearn.metrics import (
    accuracy_score,
    balanced_accuracy_score,
    classification_report,
)


def evaluate_prediction(y, y_hat):

    true_positive = 0.0
    true_negative = 0.0
    false_positive = 0.0
    false_negative = 0.0

    tp = []
    tn = []
    fp = []
    fn = []

    for i in range(len(y)):
        if y[i] == 1:
            if y_hat[i] == 1:
                true_positive += 1
                tp.append(i)
            else:
                false_negative += 1
                fn.append(i)
        else:  # -1
            if y_hat[i] == 0:
                true_negative += 1
                tn.append(i)
            else:
                false_positive += 1
                fp.append(i)

    accuracy = (true_positive + true_negative) / (
        true_positive + true_negative + false_positive + false_negative
    )

    if (true_positive + false_negative) != 0:
        sensitivity = true_positive / (true_positive + false_negative)
    else:
        sensitivity = 0.0

    if (false_positive + true_negative) != 0:
        specificity = true_negative / (false_positive + true_negative)
    else:
        specificity = 0.0

    if (true_positive + false_positive) != 0:
        ppv = true_positive / (true_positive + false_positive)
    else:
        ppv = 0.0

    if (true_negative + false_negative) != 0:
        npv = true_negative / (true_negative + false_negative)
    else:
        npv = 0.0

    balanced_accuracy = (sensitivity + specificity) / 2

    results = {
        "accuracy": accuracy,
        "balanced_accuracy": balanced_accuracy,
        "sensitivity": sensitivity,
        "specificity": specificity,
        "ppv": ppv,
        "npv": npv,
        "confusion_matrix": {
            "tp": len(tp),
            "tn": len(tn),
            "fp": len(fp),
            "fn": len(fn),
        },
    }

    return results


def gram_matrix_linear(data):
    return np.dot(data, data.transpose())


def evaluate_prediction_multiclass(y, y_hat):

    balanced_accuracy = balanced_accuracy_score(y, y_hat)
    accuracy = accuracy_score(y, y_hat)

    results = {"accuracy": accuracy, "balanced_accuracy": balanced_accuracy}

    return results


def metric_distribution(
    metric, labels, output_path, num_classes=2, metric_label="balanced accuracy"
):
    """

    Distribution plots of various metrics such as balanced accuracy!

    metric is expected to be ndarray of size [num_repetitions, num_datasets]

    """
    # from __future__ import print_function, division

    import matplotlib.pyplot as plt
    import numpy as np
    from matplotlib import cm
    from matplotlib.backends.backend_pdf import PdfPages

    num_repetitions = metric.shape[0]
    num_datasets = metric.shape[1]
    assert len(labels) == num_datasets, "Differing number of features and labels!"
    method_ticks = 1.0 + np.arange(num_datasets)

    fig, ax = plt.subplots(figsize=[9, 9])
    line_coll = ax.violinplot(
        metric,
        widths=0.8,
        bw_method=0.2,
        showmedians=True,
        showextrema=False,
        positions=method_ticks,
    )

    cmap = cm.get_cmap("Paired", num_datasets)
    for cc, ln in enumerate(line_coll["bodies"]):
        ln.set_facecolor(cmap(cc))
        ln.set_label(labels[cc])

    plt.legend(loc=2, ncol=num_datasets)

    ax.tick_params(axis="both", which="major", labelsize=15)
    ax.grid(axis="y", which="major")

    lower_lim = np.round(np.min([np.float64(0.9 / num_classes), metric.min()]), 3)
    upper_lim = np.round(np.max([1.01, metric.max()]), 3)
    step_tick = 0.1
    ax.set_ylim(lower_lim, upper_lim)

    ax.set_xticks(method_ticks)
    ax.set_xlim(np.min(method_ticks) - 1, np.max(method_ticks) + 1)
    ax.set_xticklabels(labels, rotation=45)  # 'vertical'

    ax.set_yticks(np.arange(lower_lim, upper_lim, step_tick))
    ax.set_yticklabels(np.arange(lower_lim, upper_lim, step_tick))
    # plt.xlabel(xlabel, fontsize=16)
    plt.ylabel(metric_label, fontsize=16)

    fig.tight_layout()

    pp1 = PdfPages(output_path + ".pdf")
    pp1.savefig()
    pp1.close()

    return



# coding: utf8
from typing import Optional


def convert_phase_in_radians(in_file: str, out_file: Optional[str] = None):
    """Convert phase image in radians."""
    import math
    import os
    import os.path as op

    import nibabel as nb
    import numpy as np

    assert op.isfile(in_file)

    img = nb.load(in_file)
    imin = np.amin(img.get_fdata(dtype="float32"))
    imax = np.amax(img.get_fdata(dtype="float32"))

    out_file = out_file or op.abspath("phase_in_rad.nii.gz")

    cmd = "fslmaths %s -mul %s -div %s %s -odt float" % (
        in_file,
        2.0 * math.pi,
        imax,
        out_file,
    )
    os.system(cmd)

    data = (
        img2.get_fdata(dtype="float32").astype(np.float32)
        - img1.get_fdata(dtype="float32").astype(np.float32)
    ) * (1.0 / delta_te)
    nb.Nifti1Image(data, img.get_affine(), img.get_header()).to_filename(out_file)

    return out_file


def create_phase_in_radsec(
    in_phase1, in_phase2, delta_te, out_file: Optional[str] = None
):
    """Converts input (unwarpped) phase1 and phase2 map to into a fieldmap inrads.

    Warning:
        delta_te should be in seconds.
    """
    import os.path as op

    import nibabel as nb
    import numpy as np

    out_file = out_file or op.abspath("fmap_radsec.nii.gz")

    img1 = nb.load(in_phase1)
    img2 = nb.load(in_phase2)
    data = (
        img2.get_fdata(dtype="float32").astype(np.float32)
        - img1.get_fdata(dtype="float32").astype(np.float32)
    ) * (1.0 / delta_te)
    nb.Nifti1Image(data, img1.get_affine(), img1.get_header()).to_filename(out_file)
    return out_file


def resample_fmap_to_b0(in_fmap, in_b0, out_file: Optional[str] = None):
    """Resample fieldmap onto the b0 image.

    Warnings:
        The fieldmap should already be aligned on the b0.

    Args:
        in_fmap(str): Fieldmap image.
        in_b0(str): B0 image.
        out_file(optional[str]): Filename (default: <in_fmap>_space-b0.nii.gz

    Returns:
        Fieldmap image resampled on b0.
    """
    import os.path as op

    import nibabel
    from nilearn.image import resample_to_img

    if not out_file:
        fname, ext = op.splitext(op.basename(in_fmap))
        if ext == ".gz":
            fname, ext2 = op.splitext(fname)
            ext = ext2 + ext
        out_resampled_fmap = op.abspath(f"%s_space-b0%s" % (fname, ext))
    else:
        out_resampled_fmap = out_file

    resampled_fmap = resample_to_img(
        source_img=in_fmap, target_img=in_b0, interpolation="continuous"
    )

    nibabel.nifti1.save(resampled_fmap, out_resampled_fmap)

    return out_resampled_fmap



# coding: utf8


def prepare_flowfields(flow_fields, tissues):
    return [[f] * len(tissues) for f in flow_fields]


def join_smoothed_files(smoothed_normalized_files):
    """Join outputs."""
    return [
        [x for smooth in subject for x in smooth]
        for subject in zip(*smoothed_normalized_files)
    ]



# coding: utf-8

import clinica.pipelines.engine as cpe


class PetSurface(cpe.Pipeline):
    """PetSurface - Surface-based processing of PET images.

    Returns:
        A clinica pipeline object containing the PetSurface pipeline.
    """

    def check_pipeline_parameters(self):
        """Check pipeline parameters."""
        if "acq_label" not in self.parameters.keys():
            raise KeyError("Missing compulsory acq_label key in pipeline parameter.")
        if "suvr_reference_region" not in self.parameters.keys():
            raise KeyError(
                "Missing compulsory suvr_reference_region key in pipeline parameter."
            )
        if "pvc_psf_tsv" not in self.parameters.keys():
            raise KeyError("Missing compulsory pvc_psf_tsv key in pipeline parameter.")

    def check_custom_dependencies(self):
        """Check dependencies that can not be listed in the `info.json` file."""

    def get_input_fields(self):
        """Specify the list of possible inputs of this pipeline.

        Returns:
            A list of (string) input fields name.
        """
        return [
            "orig_nu",
            "pet",
            "white_surface_left",
            "white_surface_right",
            "destrieux_left",
            "destrieux_right",
            "desikan_left",
            "desikan_right",
        ]

    def get_output_fields(self):
        """Specify the list of possible outputs of this pipeline."""
        return []

    def build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import os

        from clinica.utils.filemanip import save_participants_sessions
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        if self.parameters["longitudinal"]:
            self.build_input_node_longitudinal()
        else:
            self.build_input_node_cross_sectional()

        # Save subjects to process in <WD>/<Pipeline.name>/participants.tsv
        folder_participants_tsv = os.path.join(self.base_dir, self.name)
        save_participants_sessions(
            self.subjects, self.sessions, folder_participants_tsv
        )

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint(
                "List available in %s"
                % os.path.join(folder_participants_tsv, "participants.tsv")
            )
            cprint("The pipeline will last approximately a few hours per image.")

    def build_input_node_longitudinal(self):
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.utils.input_files as input_files
        from clinica.iotools.utils.data_handling import (
            check_relative_volume_location_in_world_coordinate_system,
        )
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.inputs import clinica_file_reader

        read_parameters_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
            synchronize=True,
        )

        all_errors = []
        try:
            read_parameters_node.inputs.pet = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.bids_directory,
                input_files.bids_pet_nii(self.parameters["acq_label"]),
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.orig_nu = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_LONG_ORIG_NU,
            )

        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.white_surface_right = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_LONG_SURF_R,
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.white_surface_left = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_LONG_SURF_L,
            )

        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.destrieux_left = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_LONG_DESTRIEUX_PARC_L,
            )

        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.destrieux_right = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_LONG_DESTRIEUX_PARC_R,
            )

        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.desikan_left = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_LONG_DESIKAN_PARC_L,
            )

        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.desikan_right = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_LONG_DESIKAN_PARC_R,
            )

        except ClinicaException as e:
            all_errors.append(e)

        if len(all_errors) > 0:
            error_message = "Clinica faced errors while trying to read files in your BIDS or CAPS directories.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaException(error_message)

        check_relative_volume_location_in_world_coordinate_system(
            "T1w-MRI (orig_nu.mgz)",
            read_parameters_node.inputs.orig_nu,
            self.parameters["acq_label"] + " PET",
            read_parameters_node.inputs.pet,
            self.bids_directory,
            self.parameters["acq_label"],
        )

        # fmt: off
        self.connect(
            [
                (read_parameters_node, self.input_node, [("pet", "pet"),
                                                         ("orig_nu", "orig_nu"),
                                                         ("white_surface_left", "white_surface_left"),
                                                         ("white_surface_right", "white_surface_right"),
                                                         ("destrieux_left", "destrieux_left"),
                                                         ("destrieux_right", "destrieux_right"),
                                                         ("desikan_left", "desikan_left"),
                                                         ("desikan_right", "desikan_right")])
            ]
        )
        # fmt: on

    def build_input_node_cross_sectional(self):
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        import clinica.utils.input_files as input_files
        from clinica.iotools.utils.data_handling import (
            check_relative_volume_location_in_world_coordinate_system,
        )
        from clinica.utils.exceptions import ClinicaException
        from clinica.utils.inputs import clinica_file_reader

        read_parameters_node = npe.Node(
            name="LoadingCLIArguments",
            interface=nutil.IdentityInterface(
                fields=self.get_input_fields(), mandatory_inputs=True
            ),
            synchronize=True,
        )

        all_errors = []
        try:

            read_parameters_node.inputs.pet = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.bids_directory,
                input_files.bids_pet_nii(self.parameters["acq_label"]),
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.orig_nu = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_ORIG_NU,
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.white_surface_right = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_WM_SURF_R,
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.white_surface_left = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_WM_SURF_L,
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.destrieux_left = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_DESTRIEUX_PARC_L,
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.destrieux_right = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_DESTRIEUX_PARC_R,
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.desikan_left = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_DESIKAN_PARC_L,
            )
        except ClinicaException as e:
            all_errors.append(e)

        try:
            read_parameters_node.inputs.desikan_right = clinica_file_reader(
                self.subjects,
                self.sessions,
                self.caps_directory,
                input_files.T1_FS_DESIKAN_PARC_R,
            )
        except ClinicaException as e:
            all_errors.append(e)

        if len(all_errors) > 0:
            error_message = "Clinica faced errors while trying to read files in your BIDS or CAPS directories.\n"
            for msg in all_errors:
                error_message += str(msg)
            raise ClinicaException(error_message)

        check_relative_volume_location_in_world_coordinate_system(
            "T1w-MRI (orig_nu.mgz)",
            read_parameters_node.inputs.orig_nu,
            self.parameters["acq_label"] + " PET",
            read_parameters_node.inputs.pet,
            self.bids_directory,
            self.parameters["acq_label"],
        )

        # fmt: off
        self.connect(
            [
                (read_parameters_node, self.input_node, [("pet", "pet"),
                                                         ("orig_nu", "orig_nu"),
                                                         ("white_surface_left", "white_surface_left"),
                                                         ("white_surface_right", "white_surface_right"),
                                                         ("destrieux_left", "destrieux_left"),
                                                         ("destrieux_right", "destrieux_right"),
                                                         ("desikan_left", "desikan_left"),
                                                         ("desikan_right", "desikan_right")])
            ]
        )
        # fmt: on

    def build_output_node(self):
        """Build and connect an output node to the pipeline."""

    def build_core_nodes(self):
        """Build and connect the core nodes of the pipeline.

        The function get_wf constructs a pipeline for one subject (in pet_surface_utils.py) and runs it.
        We use iterables to give to the node all the files and information needed.
        """
        # TODO(@arnaud.marcoux): Convert it to a Node with iterables + MapNodes.
        #   I'm experimenting something to avoid the "MapNode of MapNode" case
        #   with iterables. I'll try to apply it on the tractography pipeline.
        #   Check it out to get inspiration from it when it's ready.

        import os

        import nipype.interfaces.utility as niu
        import nipype.pipeline.engine as npe

        import clinica.pipelines.pet_surface.pet_surface_utils as utils
        from clinica.utils.spm import spm_standalone_is_available, use_spm_standalone

        full_pipe = npe.MapNode(
            niu.Function(
                input_names=[
                    "subject_id",
                    "session_id",
                    "caps_dir",
                    "pvc_psf_tsv",
                    "pet",
                    "orig_nu",
                    "white_surface_left",
                    "white_surface_right",
                    "working_directory_subjects",
                    "acq_label",
                    "csv_segmentation",
                    "suvr_reference_region",
                    "matscript_folder_inverse_deformation",
                    "desikan_left",
                    "desikan_right",
                    "destrieux_left",
                    "destrieux_right",
                    "spm_standalone_is_available",
                    "is_longitudinal",
                ],
                output_names=[],
                function=utils.get_wf,
            ),
            name="full_pipeline_mapnode",
            iterfield=[
                "subject_id",
                "session_id",
                "pet",
                "orig_nu",
                "white_surface_left",
                "white_surface_right",
                "desikan_left",
                "desikan_right",
                "destrieux_left",
                "destrieux_right",
            ],
        )

        full_pipe.inputs.subject_id = self.subjects
        full_pipe.inputs.session_id = self.sessions
        full_pipe.inputs.caps_dir = self.caps_directory
        full_pipe.inputs.pvc_psf_tsv = self.parameters["pvc_psf_tsv"]
        full_pipe.inputs.working_directory_subjects = self.base_dir
        full_pipe.inputs.acq_label = self.parameters["acq_label"]
        full_pipe.inputs.suvr_reference_region = self.parameters[
            "suvr_reference_region"
        ]
        full_pipe.inputs.csv_segmentation = os.path.abspath(
            os.path.join(
                os.path.dirname(os.path.realpath(__file__)),
                "..",
                "..",
                "resources",
                "label_conversion_gtmsegmentation.csv",
            )
        )
        full_pipe.inputs.matscript_folder_inverse_deformation = os.path.abspath(
            os.path.dirname(os.path.realpath(__file__))
        )
        full_pipe.inputs.is_longitudinal = self.parameters["longitudinal"]

        # This section of code determines whether to use SPM standalone or not
        if spm_standalone_is_available():
            use_spm_standalone()
            full_pipe.inputs.spm_standalone_is_available = True
        else:
            full_pipe.inputs.spm_standalone_is_available = False

        # Connection
        # ==========
        # fmt: off
        self.connect(
            [
                (self.input_node, full_pipe, [("pet", "pet"),
                                              ("white_surface_left", "white_surface_left"),
                                              ("white_surface_right", "white_surface_right"),
                                              ("orig_nu", "orig_nu"),
                                              ("destrieux_left", "destrieux_left"),
                                              ("destrieux_right", "destrieux_right"),
                                              ("desikan_left", "desikan_left"),
                                              ("desikan_right", "desikan_right")])
            ]
        )
        # fmt: off
